from PySide6.QtWidgets import (
    QWidget,
    QGridLayout,
    QPushButton,
    QLineEdit,
    QFileDialog,
    QStyle,
    QApplication,
    QDialog,
    QLabel,
    QDialogButtonBox,
)
from PySide6.QtCore import Signal, Qt

import os


class Path_Button_Edit(QWidget):
    """This class provides QLineEdit with a QPushButton, used to select
    a file-path, that is then displayed in the LineEdit.

    Parameters
    ----------

    Returns
    -------

    """

    path_changed = Signal(str)

    def __init__(
        self,
        parent=None,
        path="",
        default_dir="",
        file_extension="",
        select_directory=False,
        save_file=False,
    ):
        super().__init__(parent)
        layout = QGridLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.button = QPushButton("...")
        self.button.setIcon(
            QApplication.style().standardIcon(QStyle.SP_DialogOpenButton)
        )
        self.line = QLineEdit()
        layout.addWidget(self.line, 0, 0)
        layout.addWidget(self.button, 0, 1)
        if path:
            self.set_path(path)
        self.default_dir = default_dir
        self.button.clicked.connect(self.choose_path)
        self.line.textChanged.connect(self.changed)
        self.file_extension = file_extension
        self.select_directory = select_directory
        self.save_file = save_file

    def get_path(self):
        """ """
        return self.line.text()

    def set_path(self, path):
        """

        Parameters
        ----------
        path :


        Returns
        -------

        """
        self.line.setText(path)

    def choose_path(self):
        """ """
        try:
            direc = os.path.dirname(self.get_path()) or self.default_dir
        except OSError:
            direc = self.default_dir
        if self.select_directory:
            path = QFileDialog.getExistingDirectory(self, "Select Directory", dir=direc)
        elif self.save_file:
            path = QFileDialog.getSaveFileName(
                self, "Choose Filename", dir=direc, filter=self.file_extension
            )[0]
        else:
            path = QFileDialog.getOpenFileName(
                self, "Select File", dir=direc, filter=self.file_extension
            )[0]
        if path:
            self.line.setText(path)

    def changed(self):
        """ """
        self.path_changed.emit(self.get_path())


class Path_Button_Dialog(QDialog):
    def __init__(
        self,
        parent=None,
        path="",
        default_dir="",
        file_extension="",
        select_directory=False,
        save_file=False,
        title="",
        text="",
    ):
        super().__init__(parent)
        layout = QGridLayout()
        self.setLayout(layout)

        if text:
            label = QLabel(text)
            layout.addWidget(label, 0, 0)
        self.pathEdit = Path_Button_Edit(
            self, path, default_dir, file_extension, select_directory, save_file
        )
        layout.addWidget(self.pathEdit, 1, 0)
        if title:
            self.setWindowTitle(title)

        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")

        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)
        layout.addWidget(self.buttonBox, 2, 0)

        self.path = ""

    def accept(self) -> None:
        self.path = self.pathEdit.get_path()
        super().accept()



import os.path
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QLabel,
    QLineEdit,
    QGridLayout,
    QDialogButtonBox,
    QComboBox,
)

from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
from nomad_camels.nomad_integration import nomad_communication


class UploadDialog(QDialog):
    """UI widget to handle uploading data to NOMAD."""

    def __init__(self, parent=None, file_set="", upload_path=""):
        super().__init__(parent)

        label_file = QLabel("File:")
        self.pathEdit_file = Path_Button_Edit(self, file_set)

        label_upload = QLabel("Upload:")
        self.combobox_upload = QComboBox()
        uploads = nomad_communication.get_user_upload_names(self)
        self.combobox_upload.addItems(uploads)

        label_upload_path = QLabel("Directory in upload:")
        if upload_path:
            self.lineEdit_upload_path = QLineEdit(upload_path)
        else:
            self.lineEdit_upload_path = QLineEdit("CAMELS_data")

        self.button_box = QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        )
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        layout = QGridLayout()

        layout.addWidget(label_file, 10, 0)
        layout.addWidget(self.pathEdit_file, 10, 1)
        layout.addWidget(label_upload, 11, 0)
        layout.addWidget(self.combobox_upload, 11, 1)

        layout.addWidget(label_upload_path, 13, 0)
        layout.addWidget(self.lineEdit_upload_path, 13, 1)

        layout.addWidget(self.button_box, 20, 0, 1, 2)

        self.setLayout(layout)
        if file_set:
            mid = os.path.basename(file_set) + " "
            self.pathEdit_file.setHidden(True)
            label_file.setHidden(True)
        else:
            mid = ""
        self.setWindowTitle(f"Upload {mid}to NOMAD")
        self.adjustSize()
        self.exec()

    def accept(self) -> None:
        """ "Before calling the super method, the upload is done."""
        f = self.pathEdit_file.get_path()
        upload = self.combobox_upload.currentText()
        path = self.lineEdit_upload_path.text()
        nomad_communication.upload_file(f, upload, path)
        super().accept()


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    dialog = UploadDialog()
    dialog.exec()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'for_loop.ui'
##
## Created by: Qt User Interface Compiler version 6.6.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QGridLayout,
    QHeaderView, QLabel, QPushButton, QRadioButton,
    QSizePolicy, QSpacerItem, QTableWidget, QTableWidgetItem,
    QWidget)

from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box

class Ui_for_loop_config(object):
    def setupUi(self, for_loop_config):
        if not for_loop_config.objectName():
            for_loop_config.setObjectName(u"for_loop_config")
        for_loop_config.resize(380, 476)
        self.gridLayout = QGridLayout(for_loop_config)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.pushButton_add_point = QPushButton(for_loop_config)
        self.pushButton_add_point.setObjectName(u"pushButton_add_point")
        self.pushButton_add_point.setMinimumSize(QSize(30, 23))
        self.pushButton_add_point.setMaximumSize(QSize(30, 16777215))

        self.gridLayout.addWidget(self.pushButton_add_point, 1, 3, 1, 1)

        self.label_9 = QLabel(for_loop_config)
        self.label_9.setObjectName(u"label_9")
        self.label_9.setMaximumSize(QSize(80, 16777215))

        self.gridLayout.addWidget(self.label_9, 1, 2, 1, 1)

        self.pushButton_del_point = QPushButton(for_loop_config)
        self.pushButton_del_point.setObjectName(u"pushButton_del_point")
        self.pushButton_del_point.setMinimumSize(QSize(30, 23))
        self.pushButton_del_point.setMaximumSize(QSize(30, 16777215))

        self.gridLayout.addWidget(self.pushButton_del_point, 1, 4, 1, 1)

        self.comboBox_loop_type = QComboBox(for_loop_config)
        self.comboBox_loop_type.addItem("")
        self.comboBox_loop_type.addItem("")
        self.comboBox_loop_type.addItem("")
        self.comboBox_loop_type.addItem("")
        self.comboBox_loop_type.addItem("")
        self.comboBox_loop_type.setObjectName(u"comboBox_loop_type")
        self.comboBox_loop_type.setMinimumSize(QSize(135, 0))

        self.gridLayout.addWidget(self.comboBox_loop_type, 1, 1, 1, 1)

        self.label = QLabel(for_loop_config)
        self.label.setObjectName(u"label")

        self.gridLayout.addWidget(self.label, 1, 0, 1, 1)

        self.path_line_button = Path_Button_Edit(for_loop_config)
        self.path_line_button.setObjectName(u"path_line_button")
        self.path_line_button.setMinimumSize(QSize(0, 30))

        self.gridLayout.addWidget(self.path_line_button, 5, 0, 1, 2)

        self.sweep_widget = QWidget(for_loop_config)
        self.sweep_widget.setObjectName(u"sweep_widget")
        self.gridLayout_2 = QGridLayout(self.sweep_widget)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.label_max = QLabel(self.sweep_widget)
        self.label_max.setObjectName(u"label_max")

        self.gridLayout_2.addWidget(self.label_max, 5, 1, 1, 1)

        self.verticalSpacer_3 = QSpacerItem(20, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer_3, 0, 0, 1, 1)

        self.lineEdit_max = Variable_Box(self.sweep_widget)
        self.lineEdit_max.setObjectName(u"lineEdit_max")

        self.gridLayout_2.addWidget(self.lineEdit_max, 6, 1, 1, 1)

        self.lineEdit_point_distance = Variable_Box(self.sweep_widget)
        self.lineEdit_point_distance.setObjectName(u"lineEdit_point_distance")
        self.lineEdit_point_distance.setEnabled(True)

        self.gridLayout_2.addWidget(self.lineEdit_point_distance, 9, 1, 1, 1)

        self.label_3 = QLabel(self.sweep_widget)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout_2.addWidget(self.label_3, 3, 1, 1, 1)

        self.lineEdit_stop = Variable_Box(self.sweep_widget)
        self.lineEdit_stop.setObjectName(u"lineEdit_stop")

        self.gridLayout_2.addWidget(self.lineEdit_stop, 4, 1, 1, 1)

        self.label_min = QLabel(self.sweep_widget)
        self.label_min.setObjectName(u"label_min")

        self.gridLayout_2.addWidget(self.label_min, 5, 0, 1, 1)

        self.label_6 = QLabel(self.sweep_widget)
        self.label_6.setObjectName(u"label_6")

        self.gridLayout_2.addWidget(self.label_6, 1, 0, 1, 1)

        self.label_2 = QLabel(self.sweep_widget)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout_2.addWidget(self.label_2, 3, 0, 1, 1)

        self.lineEdit_start = Variable_Box(self.sweep_widget)
        self.lineEdit_start.setObjectName(u"lineEdit_start")

        self.gridLayout_2.addWidget(self.lineEdit_start, 4, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer, 2, 0, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer_2, 11, 0, 1, 1)

        self.comboBox_sweep_mode = QComboBox(self.sweep_widget)
        self.comboBox_sweep_mode.addItem("")
        self.comboBox_sweep_mode.addItem("")
        self.comboBox_sweep_mode.addItem("")
        self.comboBox_sweep_mode.addItem("")
        self.comboBox_sweep_mode.setObjectName(u"comboBox_sweep_mode")

        self.gridLayout_2.addWidget(self.comboBox_sweep_mode, 1, 1, 1, 1)

        self.lineEdit_min = Variable_Box(self.sweep_widget)
        self.lineEdit_min.setObjectName(u"lineEdit_min")

        self.gridLayout_2.addWidget(self.lineEdit_min, 6, 0, 1, 1)

        self.lineEdit_n_points = Variable_Box(self.sweep_widget)
        self.lineEdit_n_points.setObjectName(u"lineEdit_n_points")

        self.gridLayout_2.addWidget(self.lineEdit_n_points, 9, 0, 1, 1)

        self.checkBox_include_endpoints = QCheckBox(self.sweep_widget)
        self.checkBox_include_endpoints.setObjectName(u"checkBox_include_endpoints")
        self.checkBox_include_endpoints.setChecked(True)

        self.gridLayout_2.addWidget(self.checkBox_include_endpoints, 10, 0, 1, 2)

        self.radioButton_n_points = QRadioButton(self.sweep_widget)
        self.radioButton_n_points.setObjectName(u"radioButton_n_points")

        self.gridLayout_2.addWidget(self.radioButton_n_points, 7, 0, 1, 1)

        self.radioButton_point_distance = QRadioButton(self.sweep_widget)
        self.radioButton_point_distance.setObjectName(u"radioButton_point_distance")
        self.radioButton_point_distance.setChecked(True)

        self.gridLayout_2.addWidget(self.radioButton_point_distance, 7, 1, 1, 1)


        self.gridLayout.addWidget(self.sweep_widget, 3, 0, 1, 2)

        self.verticalSpacer_4 = QSpacerItem(20, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_4, 6, 1, 1, 1)

        self.tableWidget_points = QTableWidget(for_loop_config)
        self.tableWidget_points.setObjectName(u"tableWidget_points")
        self.tableWidget_points.setMaximumSize(QSize(150, 16777215))

        self.gridLayout.addWidget(self.tableWidget_points, 2, 2, 5, 3)

        QWidget.setTabOrder(self.comboBox_loop_type, self.pushButton_add_point)
        QWidget.setTabOrder(self.pushButton_add_point, self.pushButton_del_point)
        QWidget.setTabOrder(self.pushButton_del_point, self.comboBox_sweep_mode)
        QWidget.setTabOrder(self.comboBox_sweep_mode, self.lineEdit_start)
        QWidget.setTabOrder(self.lineEdit_start, self.lineEdit_stop)
        QWidget.setTabOrder(self.lineEdit_stop, self.lineEdit_min)
        QWidget.setTabOrder(self.lineEdit_min, self.lineEdit_max)
        QWidget.setTabOrder(self.lineEdit_max, self.lineEdit_n_points)
        QWidget.setTabOrder(self.lineEdit_n_points, self.lineEdit_point_distance)
        QWidget.setTabOrder(self.lineEdit_point_distance, self.checkBox_include_endpoints)
        QWidget.setTabOrder(self.checkBox_include_endpoints, self.tableWidget_points)

        self.retranslateUi(for_loop_config)

        QMetaObject.connectSlotsByName(for_loop_config)
    # setupUi

    def retranslateUi(self, for_loop_config):
        for_loop_config.setWindowTitle(QCoreApplication.translate("for_loop_config", u"Form", None))
        self.pushButton_add_point.setText(QCoreApplication.translate("for_loop_config", u"+", None))
        self.label_9.setText(QCoreApplication.translate("for_loop_config", u"Points:", None))
        self.pushButton_del_point.setText(QCoreApplication.translate("for_loop_config", u"-", None))
        self.comboBox_loop_type.setItemText(0, QCoreApplication.translate("for_loop_config", u"start - stop", None))
        self.comboBox_loop_type.setItemText(1, QCoreApplication.translate("for_loop_config", u"start - min - max - stop", None))
        self.comboBox_loop_type.setItemText(2, QCoreApplication.translate("for_loop_config", u"start - max - min - stop", None))
        self.comboBox_loop_type.setItemText(3, QCoreApplication.translate("for_loop_config", u"Value-List", None))
        self.comboBox_loop_type.setItemText(4, QCoreApplication.translate("for_loop_config", u"Text-File", None))

        self.label.setText(QCoreApplication.translate("for_loop_config", u"Loop-Type:", None))
        self.label_max.setText(QCoreApplication.translate("for_loop_config", u"Max", None))
        self.label_3.setText(QCoreApplication.translate("for_loop_config", u"Stop", None))
        self.label_min.setText(QCoreApplication.translate("for_loop_config", u"Min", None))
        self.label_6.setText(QCoreApplication.translate("for_loop_config", u"Sweep mode", None))
        self.label_2.setText(QCoreApplication.translate("for_loop_config", u"Start", None))
        self.comboBox_sweep_mode.setItemText(0, QCoreApplication.translate("for_loop_config", u"linear", None))
        self.comboBox_sweep_mode.setItemText(1, QCoreApplication.translate("for_loop_config", u"logarithmic", None))
        self.comboBox_sweep_mode.setItemText(2, QCoreApplication.translate("for_loop_config", u"exponential", None))
        self.comboBox_sweep_mode.setItemText(3, QCoreApplication.translate("for_loop_config", u"1/x", None))

        self.checkBox_include_endpoints.setText(QCoreApplication.translate("for_loop_config", u"Include end-points", None))
        self.radioButton_n_points.setText(QCoreApplication.translate("for_loop_config", u"# points", None))
        self.radioButton_point_distance.setText(QCoreApplication.translate("for_loop_config", u"point distance", None))
    # retranslateUi




"""Provides helping functions to load and save all settings inside CAMELS.

Attributes
----------
    appdata_path : str, path
        The path to the local appdata of NOMAD CAMELS
    preset_path : str, path
        the path, where CAMELS saves the presets, a subfolder of `appdata_path`
    backup_path : str, path
        the path, where CAMELS saves backups of presets, a subfolder of
        `preset_path`
    save_string_list : list[class]
        a list of QWidget-classes, for which the current string they display should be
        saved
    save_dict_skip : list[class]
        a list of QWidget-classes, that should be ignored when saving
    standard_pref : dict
        dictionary of the default preferences used if nothing is changed by the
        user
"""

import os.path
import sys
from os.path import isdir
from os import makedirs, getenv, listdir
from shutil import copyfile
import importlib
from glob import glob
import platform

import numpy as np
import pandas as pd
from PySide6.QtWidgets import (
    QComboBox,
    QLineEdit,
    QWidget,
    QSplitter,
    QLabel,
    QPushButton,
    QTreeView,
    QListView,
    QMenuBar,
    QMenu,
    QStatusBar,
    QGridLayout,
)
from PySide6.QtGui import QAction
from PySide6.QtCore import QTimer, Signal, SignalInstance

from datetime import datetime, timedelta
import json
import ophyd

from nomad_camels.main_classes import protocol_class, device_class, loop_step
from nomad_camels.utility.load_save_helper_functions import load_plots
from nomad_camels.utility.device_handling import load_local_packages
from nomad_camels.utility import variables_handling
from nomad_camels.ui_widgets.warn_popup import WarnPopup
from nomad_camels.bluesky_handling.watchdogs import Watchdog
from nomad_camels.bluesky_handling.evaluation_helper import Evaluator


os_name = platform.system()

if os_name == "Windows":
    # Use the APPDATA environment variable on Windows
    data_path = os.path.join(getenv("USERPROFILE"), "Documents", "NOMAD_CAMELS_data")
    appdata_path = os.path.join(getenv("LOCALAPPDATA"), "nomad_camels")
elif os_name == "Linux":
    # Use the XDG_DATA_HOME environment variable on Linux, defaulting to ~/.local/share
    data_path = os.path.join(
        getenv("XDG_DATA_HOME")
        or os.path.join(os.path.expanduser("~"), ".local", "share"),
        "nomad_camels_data",
    )
    appdata_path = os.path.join(
        getenv("XDG_CONFIG_HOME") or os.path.join(os.path.expanduser("~"), ".config"),
        "nomad_camels",
    )
elif os_name == "Darwin":
    # Use ~/Library/Application Support on MacOS
    data_path = os.path.join(
        os.path.expanduser("~"), "Library", "Application Support", "nomad_camels_data"
    )
    appdata_path = os.path.join(
        os.path.expanduser("~"), "Library", "Application Support", "nomad_camels"
    )
else:
    # Default to the home directory
    data_path = os.path.expanduser("~", "nomad_camels_data")
    appdata_path = os.path.expanduser("~", ".nomad_camels")


preset_path = os.path.join(appdata_path, "Presets")
backup_path = os.path.join(preset_path, "Backup")
save_string_list = [QComboBox, QLineEdit, QTreeView, QListView]
save_dict_skip = [
    QWidget,
    QSplitter,
    QLabel,
    QPushButton,
    QMenu,
    QMenuBar,
    QAction,
    QStatusBar,
    QGridLayout,
    QTimer,
    Evaluator,
    Signal,
    SignalInstance,
]

# Get the current operating system


standard_pref = {
    "autosave": True,
    "autosave_run": True,
    "backup_before_run": True,
    "dark_mode": False,
    "graphic_theme": "Fusion",
    "n_decimals": 3,
    "number_format": "mixed",
    "mixed_from": 3,
    "py_files_path": os.path.join(appdata_path, "python_files").replace("\\", "/"),
    "meas_files_path": data_path,
    "device_driver_path": os.path.join(
        os.getcwd(), "devices", "devices_drivers"
    ).replace("\\", "/"),
    "databroker_catalog_name": "CAMELS_CATALOG",
    "driver_repository": "https://github.com/FAU-LAP/CAMELS_drivers",
    "repo_branch": "main",
    "repo_directory": "",
    "play_camel_on_error": False,
    "auto_check_updates": False,
    "log_level": "Warning",
    "logfile_size": 1,
    "logfile_backups": 1,
    "NOMAD_URL": "",
    "backups": "smart",
    "backup_number": 30,
    "number_databroker_files": 30,
    "extension_path": os.path.join(os.getcwd(), "extensions").replace("\\", "/"),
    "extensions": [],
    "password_protection": False,
    "password_hash": "",
    "new_file_each_run": True,
    "enable_API": False,
    "API_port": "5000",
    "last_shown_notes": "0.0.0",
}


def update_config_path(path):
    global appdata_path, preset_path, backup_path
    appdata_path = path
    with open(
        os.path.join(os.path.dirname(os.path.dirname(__file__)), "config_path"), "w"
    ) as f:
        f.write(appdata_path)
    preset_path = os.path.join(appdata_path, "Presets")
    backup_path = os.path.join(preset_path, "Backup")
    standard_pref["py_files_path"] = os.path.join(appdata_path, "python_files").replace(
        "\\", "/"
    )
    if not os.path.isdir(appdata_path):
        makedirs(appdata_path)


def check_config_path():
    global appdata_path
    config_path_file = os.path.join(
        os.path.dirname(os.path.dirname(__file__)), "config_path"
    )
    if os.path.isfile(config_path_file):
        with open(config_path_file, "r") as f:
            appdata_path = f.read().strip()
    else:
        from nomad_camels.ui_widgets.path_button_edit import Path_Button_Dialog

        dialog = Path_Button_Dialog(
            path=appdata_path,
            default_dir=os.path.dirname(appdata_path),
            select_directory=True,
            title="Select path for configuration files",
            text="Please select where NOMAD CAMELS should save its configuration files",
        )
        if dialog.exec():
            appdata_path = dialog.path
        else:
            WarnPopup(
                text="No path selected, a default path will be used.\nYou can change the path in the settings.",
                title="No path selected",
                info_icon=True,
            )
    update_config_path(appdata_path)


def get_preset_list():
    """
    DEPRECATED

    returns a two list of available presets, once for devices, once
    for measurements. (files with ".predev" or ".premeas" in
    appdata_path. If the directory does not exist, it is created."""
    if isdir(preset_path):
        names = listdir(preset_path)
        if "Backup" not in names:
            makedirs(os.path.join(preset_path, "Backup"))
        presets = []
        for name in names:
            if name.endswith(".preset"):
                presets.append(name[:-7])
        return sorted(presets, key=lambda x: x.lower())
    else:
        makedirs(preset_path)
        return get_preset_list()


def autosave_preset(preset: str, preset_data, do_backup=True):
    """Saves the given preset and makes a backup of the former one in
    the backup-folder (if do_backup).

    Parameters
    ----------
    preset : str
        the name of the preset
    preset_data : dict
        data of the preset, formatted into a dictionary
    do_backup : bool, default=True
        if True, the old preset file will be copied into the backup folder, with
        timestamp
    """
    preset_file = f"{preset}.preset"
    if not os.path.isdir(preset_path):
        makedirs(preset_path)
    with open(
        os.path.join(preset_path, preset_file), "w", encoding="utf-8"
    ) as json_file:
        json.dump(preset_data, json_file, indent=2)
    if do_backup:
        make_backup(preset_file)


def save_preset(path, preset_data: dict):
    """Saves the given preset_data under the specified path.
    Further, `autosave_preset` is called.

    Parameters
    ----------
    path : str, path
        where to save the preset
    preset_data : dict
        the data of the preset, to be dumped as json
    """
    preset_name = path.split("/")[-1][:-7]
    with open(path, "w", encoding="utf-8") as json_file:
        json.dump(preset_data, json_file, indent=2)
    autosave_preset(preset_name, preset_data)


def save_dictionary(path, dictionary: dict):
    """
    Processes the given `dictionary` using `get_save_str`, then saves it as json
    to the given `path`.

    Parameters
    ----------
    path : str, path
        where to save the dictionary
    dictionary : dict
        the dictionary to be saved as json
    """
    save_dict = {}
    for key, val in dictionary.items():
        add_string = get_save_str(val)
        if add_string is not None:
            save_dict[key] = add_string
    if not os.path.isdir(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    with open(path, "w", encoding="utf-8") as file:
        json.dump(save_dict, file, indent=2)


def make_backup(preset_file: str):
    """
    Puts a copy of the given `preset_file` into the backup-folder of
    the preset. The current datetime is added to the filename.

    Parameters
    ----------
    preset_file : str
        The name of the preset file. The file needs to be in the `preset_path`.
    """
    backup_save_path = os.path.join(backup_path, preset_file[:-7])
    if not isdir(backup_save_path):
        makedirs(backup_save_path)
    now = datetime.now()
    backup_name = os.path.join(
        backup_save_path, f'{now.strftime("%Y-%m-%d_%H-%M-%S")}_{preset_file}'
    )
    copyfile(os.path.join(preset_path, preset_file), backup_name)

    if not "backups" in variables_handling.preferences:
        return
    backups = sorted(glob(f"{backup_save_path}/*.preset"), key=os.path.getmtime)
    if variables_handling.preferences["backups"] == "number":
        if "backup_number" in variables_handling.preferences:
            backup_number = variables_handling.preferences["backup_number"]
            while len(backups) > backup_number:
                os.remove(backups.pop(0))
    elif variables_handling.preferences["backups"] == "smart":
        backups = [
            b
            for b in backups
            if datetime.fromtimestamp(os.path.getmtime(b)) < now - timedelta(days=7)
        ]
        for day in range(8, 31):
            day_ago = now - timedelta(days=day)
            daily_backups = [
                b
                for b in backups
                if datetime.fromtimestamp(os.path.getmtime(b)).date() == day_ago.date()
            ]
            while len(daily_backups) > 1:
                backup = daily_backups.pop(0)
                os.remove(backup)
                backups.remove(backup)
        backups = [
            b
            for b in backups
            if datetime.fromtimestamp(os.path.getmtime(b)) < now - timedelta(days=30)
        ]
        for month in range(1, 13):
            month_ago = now - timedelta(days=30 * month)
            monthly_backups = [
                b
                for b in backups
                if datetime.fromtimestamp(os.path.getmtime(b)).month == month_ago.month
            ]
            while len(monthly_backups) > 1:
                backup = monthly_backups.pop(0)
                os.remove(backup)
                backups.remove(backup)
        backups = [
            b
            for b in backups
            if datetime.fromtimestamp(os.path.getmtime(b)) < now - timedelta(days=365)
        ]
        if not backups:
            return
        for year in range(
            1, now.year - datetime.fromtimestamp(os.path.getmtime(backups[0])).year + 1
        ):
            year_ago = now - timedelta(days=365 * year)
            yearly_backups = [
                b
                for b in backups
                if datetime.fromtimestamp(os.path.getmtime(b)).year == year_ago.year
            ]
            while len(yearly_backups) > 1:
                backup = yearly_backups.pop(0)
                os.remove(backup)
                backups.remove(backup)


def load_save_dict(
    string_dict: dict,
    object_dict: dict,
    update_missing_key=False,
    remove_extra_key=False,
):
    """For all keys both given dictionaries have in common, the value of
    the object in object_dict will be updated to the corresponding value
    of the string in string_dict. Depending on the type of the objects already
    in `object_dict`, the values may be processed differently.
    The different possibilities are:

    - QComboBox: the value will be set as current text
    - QLineEdit: the value will be set as text
    - key=="protocols_dict": calls `load_protocols_dict`
    - key in ["active_devices_dict", "active_instruments"]: calls `load_devices_dict`
    - dict or hasattr "__save_dict__" or "__dict__": calls `load_save_dict` recursively
    - list: appends all the values in the list of the strings


    Parameters
    ----------
    string_dict : dict
        A dictionary containing mostly strings as loaded from a json as values.
    object_dict : dict
        The dictionary to be loaded into.
    update_missing_key : bool
        (Default value = False)
        If True, keys that are only in `string_dict`, but not in `object_dict`
        will be added as well.
    remove_extra_key : bool
        (Default value = False)
        If True, keys that are only in `object_dict`, but not in `string_dict`
        will be removed.
    """
    for key in string_dict:
        if key in object_dict:
            obj = object_dict[key]
            val = string_dict[key]
            if issubclass(type(obj), QComboBox):
                obj.setCurrentText(val)
            elif issubclass(type(obj), QLineEdit):
                obj.setText(val)
            elif key == "watchdogs":
                for w in val:
                    obj[w] = Watchdog(**val[w])
            elif key == "protocols_dict":
                load_protocols_dict(val, obj)
            elif key in ["active_devices_dict", "active_instruments"]:
                load_devices_dict(val, obj)
            elif isinstance(obj, dict):
                load_save_dict(val, obj, True, True)
            elif hasattr(obj, "__save_dict__") or hasattr(obj, "__dict__"):
                load_save_dict(val, obj.__dict__)
            elif type(obj) is list:
                obj.clear()
                for v in val:
                    obj.append(v)
        elif update_missing_key:
            object_dict.update({key: string_dict[key]})
    if remove_extra_key:
        rem_keys = []
        for key in object_dict:
            if key not in string_dict:
                rem_keys.append(key)
        for key in rem_keys:
            object_dict.pop(key)


def get_save_str(obj):
    """Utility function to create the string with which to save the
    object-data.
    If the object has the attribute __save_dict__, it is the return value.
    Objects of the types specified in save_dict_skip return None.
    QComboBox and QLineEdit return their current text.
    If `obj` is a dictionary or has the attribute `__dict__`, the values of the
    dictionary will be recursively read by `get_save_str`.
    Types str, bool and numbers are not converted.
    Arrays are converted to lists.

    Parameters
    ----------
    obj :
        The object that should be represented as string.

    Returns
    -------
    the string-representation of `obj`

    """
    if hasattr(obj, "__save_dict__"):
        make_save_dict(obj)
        return obj.__save_dict__
    if type(obj) in save_dict_skip:
        return None
    if isinstance(obj, ophyd.Device):
        return None
    if type(obj) in save_string_list:
        if issubclass(type(obj), QComboBox):
            return obj.currentText()
        if issubclass(type(obj), QLineEdit):
            return obj.text()
        return None
    if type(obj) is pd.DataFrame:
        obj.to_dict("list")
    if hasattr(obj, "__dict__") or type(obj) is dict:
        savedic = {}
        if isinstance(obj, dict):
            dictionary = obj
        else:
            dictionary = obj.__dict__
        for key in dictionary:
            if key == "py_package":
                continue
            savedic.update({key: get_save_str(dictionary[key])})
        return savedic
    if type(obj) in [int, float, bool, np.float64]:
        return obj
    if type(obj) is np.ndarray:
        obj = list(obj)
    if type(obj) is list:
        obj_list = []
        for p in obj:
            obj_list.append(get_save_str(p))
        return obj_list
    return str(obj)


def make_save_dict(obj):
    """Utility function to update the __save_dict__ of the given obj.
    Goes through all the keys in __dict__ and calls get_save_str on the
    object. Thus working recursively if an attribute of obj also has a
    `__save_dict__`.

    Parameters
    ----------
    obj :
        The object of which the `__save_dict__` should be updated.
    """
    for key in obj.__dict__:
        if key == "__save_dict__" or (
            isinstance(obj, device_class.Device)
            and key
            in [
                "controls",
                "ophyd_class",
                "ophyd_class_no_epics",
                "channels",
                "virtual",
                "tags",
                "files",
                "directory",
                "requirements",
                "ophyd_class_name",
                "connection",
            ]
        ):
            continue
        if key == "protocol" and isinstance(obj, loop_step.Loop_Step):
            continue
        add_string = get_save_str(obj.__dict__[key])
        if add_string is not None:
            obj.__save_dict__.update({key: get_save_str(obj.__dict__[key])})


def load_protocol(path):
    """
    Loads the protocol-json from the given `path` and converts it to a CAMELS
    Measurement_Protocol instance.

    Parameters
    ----------
    path : str, path
        the path, where protocol file can be found

    Returns
    -------
    Measurement_Protocol
        the protocol constructed from the data in the loaded file
    """
    prot_name = os.path.basename(path)[:-6]
    if not os.path.isfile(path):
        return
    with open(path, "r", encoding="utf-8") as f:
        prot_dict = json.load(f)
    prot_string_dict = {prot_name: prot_dict}
    sub_protocol = {}
    load_protocols_dict(prot_string_dict, sub_protocol)
    return sub_protocol[prot_name]


# def load_manuals_dict(string_dict, manuals_dict):
#     manuals_dict.clear()
#     for key, val in string_dict.items():
#         control_type = val['control_type']
#         control = get_manual_controls.get_control_by_type_name(control_type)


def load_protocols_dict(string_dict, prot_dict):
    """
    Specific function to construct protocols from a json string-dictionary.
    Loads the attributes of a Measurement_Protocol from the corresponding keys.

    Parameters
    ----------
    string_dict : dict{"<protocol_name>": dict}
        the dictionary with the information about the protocols (which is are
        dictinaries again)
    prot_dict : dict
        dictionary of the finished protocols
    """
    prot_dict.clear()
    for key in string_dict:
        prot_data = string_dict[key]
        prot = protocol_class.Measurement_Protocol()
        prot.name = key
        if "loop_steps" in prot_data:
            prot.load_loop_steps(prot_data["loop_steps"])
        if "plots" in prot_data:
            prot.plots = load_plots([], prot_data["plots"])
        if "filename" in prot_data:
            prot.filename = prot_data["filename"]
        if "variables" in prot_data:
            prot.variables = prot_data["variables"]
        if "metadata" in prot_data:
            prot.metadata = prot_data["metadata"]
        if "channel_metadata" in prot_data:
            prot.channel_metadata = prot_data["channel_metadata"]
        if "config_metadata" in prot_data:
            prot.config_metadata = prot_data["config_metadata"]
        if "use_nexus" in prot_data:
            prot.use_nexus = prot_data["use_nexus"]
        if "description" in prot_data:
            prot.description = prot_data["description"]
        if "export_json" in prot_data:
            prot.export_json = prot_data["export_json"]
        if "export_csv" in prot_data:
            prot.export_csv = prot_data["export_csv"]
        if "h5_during_run" in prot_data:
            prot.h5_during_run = prot_data["h5_during_run"]
        if "use_end_protocol" in prot_data:
            prot.use_end_protocol = prot_data["use_end_protocol"]
        if "end_protocol" in prot_data:
            prot.end_protocol = prot_data["end_protocol"]
        if "live_variable_update" in prot_data:
            prot.live_variable_update = prot_data["live_variable_update"]
        if "allow_live_comments" in prot_data:
            prot.allow_live_comments = prot_data["allow_live_comments"]
        prot_dict.update({key: prot})


def load_devices_dict(string_dict, devices_dict):
    """Specific function to load the dictionary of devices/instruments.
    Uses the instruments' name to import the corresponding module and create the
    instance of the instrument.

    Parameters
    ----------
    string_dict : dict{"<name>": dict}
        containing the names and information of the instruments
    devices_dict : dict
        the dictionary, where to put the loaded instruments
    """
    devices_dict.clear()
    local_packages = load_local_packages()

    path = f"{os.path.dirname(os.path.dirname(__file__))}/manual_controls/set_panel"
    sys.path.append(path)
    for key in string_dict:
        dev_data = string_dict[key]
        name = dev_data["name"]
        if name in local_packages:
            dev_lib = local_packages[name]
        else:
            try:
                dev_lib = importlib.import_module(f"nomad_camels_driver_{name}.{name}")
            except Exception as e:
                try:
                    dev_lib = importlib.import_module(f"{name}.{name}")
                except Exception as e2:
                    WarnPopup(
                        text=f'Could not import instrument module "{name}"\n{e}\n{e2}',
                        title="instrument import failed",
                    )
                    continue
                    # raise Exception(f'Could not import device module {name}\n{e}\n{e2}')
        dev = dev_lib.subclass()
        dev.name = name
        if "connection" in dev_data:
            dev.connection = dev_data["connection"]
        if "virtual" in dev_data:
            dev.virtual = dev_data["virtual"]
        if "tags" in dev_data:
            dev.tags = dev_data["tags"]
        if "files" in dev_data:
            dev.files = dev_data["files"]
        if "directory" in dev_data:
            dev.directory = dev_data["directory"]
        if "requirements" in dev_data:
            dev.requirements = dev_data["requirements"]
        if "settings" in dev_data:
            dev.settings = dev_data["settings"] or dev.settings
        if "config" in dev_data:
            dev.config = dev_data["config"] or dev.config
        if "custom_name" in dev_data:
            dev.custom_name = dev_data["custom_name"]
        if "additional_info" in dev_data:
            dev.additional_info = dev_data["additional_info"]
        devices_dict.update({key: dev})


def get_most_recent_presets(return_all=False):
    """Goes through all files in the `preset_path` and returns the newest preset.

    Returns
    -------
    preset : str
        the name of the newest preset
    """
    presets = []
    if not os.path.isdir(preset_path):
        makedirs(preset_path)
    for name in listdir(preset_path):
        if name.endswith(".preset"):
            presets.append(name)
    if preset_path is None:
        return None if not return_all else [None]
    if presets:
        presets = [
            x[:-7]
            for x in sorted(
                presets, key=lambda x: os.path.getmtime(os.path.join(preset_path, x))
            )
        ]
    else:
        return None if not return_all else [None]
    return presets[-1] if not return_all else presets


def get_preferences():
    """If a file 'preferences.json' exists in the `appdata_path`, its content
    will be loaded and returned, if no file exists, it will be created
    with an empty dictionary. Keys that are in the `standard_pref`, but not in
    the loaded preferences will be added with their default value.

    Returns
    -------
    prefs : dict
        the loaded preferences dictionary
    """
    check_config_path()
    pref_file = os.path.join(appdata_path, "preferences.json")
    if "preferences.json" not in os.listdir(appdata_path):
        with open(pref_file, "w", encoding="utf-8") as file:
            from nomad_camels.ui_widgets.path_button_edit import Path_Button_Dialog

            dialog = Path_Button_Dialog(
                path=data_path,
                default_dir=os.path.dirname(data_path),
                select_directory=True,
                title="Select data path",
                text="Please select the directory where the measurement data should be saved",
            )
            if dialog.exec():
                global standard_pref
                standard_pref["meas_files_path"] = dialog.path
            else:
                WarnPopup(
                    text="No data path selected, a default path will be used.\nYou can change the path in the settings.",
                    title="No data path selected",
                    info_icon=True,
                )
            json.dump(standard_pref, file, indent=2)
    with open(pref_file, "r", encoding="utf-8") as file:
        prefs = json.load(file)
    for key, value in standard_pref.items():
        if key not in prefs:
            prefs[key] = value
    return prefs


def save_preferences(prefs: dict):
    """Saves the given dictionary prefs as 'preferences.json' in the `appdata_path`

    Parameters
    ----------
    prefs : dict
        the preferences dictionary to be saved as json
    """
    with open(
        os.path.join(appdata_path, "preferences.json"), "w", encoding="utf-8"
    ) as file:
        json.dump(prefs, file, indent=2)



import pytest
from PySide6.QtWidgets import QApplication, QTableWidget, QWidget
from PySide6.QtTest import QSignalSpy

from nomad_camels.main_classes.list_plot import (
    Values_List_Plot,
    Live_List,
    Teleporter,
    handle_teleport,
)


@pytest.fixture(scope="module")
def app():
    """Fixture for creating a Qt application."""
    app = QApplication.instance()
    if app is None:
        app = QApplication([])
    return app


@pytest.fixture
def value_list():
    return ["value1", "value2", "value3"]


@pytest.fixture
def table():
    return QTableWidget()


@pytest.fixture
def live_list(value_list, table):
    return Live_List(value_list, table)


def test_values_list_plot_initialization(app, value_list):
    widget = Values_List_Plot(value_list)
    assert isinstance(widget, QWidget)
    assert widget.windowTitle() == "value1 ..."
    assert widget.table.rowCount() == len(value_list)


def test_values_list_plot_show_again(app, value_list, qtbot):
    widget = Values_List_Plot(value_list)
    widget.hide()
    spy = QSignalSpy(widget.reopened)
    widget.show_again()
    assert widget.isVisible()
    assert spy.count() == 1


def test_live_list_initialization(live_list, value_list):
    assert isinstance(live_list, Live_List)
    assert live_list.value_list == value_list
    assert live_list.table.rowCount() == len(value_list)
    assert live_list.table.columnCount() == 2


def test_live_list_event_handling(live_list, qtbot):
    doc = {
        "descriptor": "test_descriptor",
        "data": {"value1": 1.23, "value2": 4.56, "value3": 7.89},
        "time": 1234567890.0,
    }
    live_list.desc = "test_descriptor"
    live_list.event(doc)
    for i, val in enumerate(live_list.value_list):
        assert live_list.val_items[i].text() == f"{doc['data'][val]:7e}"


def test_live_list_add_to_table(live_list):
    live_list.add_to_table("new_value")
    assert live_list.table.rowCount() == len(live_list.value_list) + 1
    assert live_list.table.item(len(live_list.value_list), 0).text() == "new_value"


def test_live_list_descriptor(live_list):
    doc = {"name": "primary", "uid": "test_uid"}
    live_list.descriptor(doc)
    assert live_list.desc == "test_uid"


def test_live_list_start(live_list):
    doc = {"time": 1234567890.0}
    live_list.start(doc)
    assert live_list.epoch_offset == doc["time"]


def test_teleporter():
    teleporter = Teleporter()
    spy = QSignalSpy(teleporter.name_doc_escape)
    teleporter.name_doc_escape.emit("test_name", {}, None)
    assert spy.count() == 1


def test_handle_teleport():
    class MockObj:
        def __call__(self, name, doc, escape):
            self.name = name
            self.doc = doc
            self.escape = escape

    mock_obj = MockObj()
    handle_teleport("test_name", {}, mock_obj)
    assert mock_obj.name == "test_name"
    assert mock_obj.doc == {}
    assert mock_obj.escape


def test_live_list_call(live_list, qtbot):
    doc = {
        "descriptor": "test_descriptor",
        "data": {"value1": 1.23, "value2": 4.56, "value3": 7.89},
        "time": 1234567890.0,
    }
    live_list.desc = "test_descriptor"
    spy = QSignalSpy(live_list.new_data)
    live_list.__call__("event", doc)
    assert spy.count() == 1



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'settings_window.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QApplication, QCheckBox, QComboBox,
    QDialog, QDialogButtonBox, QFrame, QGridLayout,
    QHBoxLayout, QLabel, QLineEdit, QPushButton,
    QRadioButton, QSizePolicy, QSpacerItem, QSpinBox,
    QWidget)

from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
import settings_window_resource_rc

class Ui_settings_window(object):
    def setupUi(self, settings_window):
        if not settings_window.objectName():
            settings_window.setObjectName(u"settings_window")
        settings_window.resize(670, 606)
        self.gridLayout = QGridLayout(settings_window)
        self.gridLayout.setObjectName(u"gridLayout")
        self.line_vertical = QFrame(settings_window)
        self.line_vertical.setObjectName(u"line_vertical")
        self.line_vertical.setFrameShadow(QFrame.Shadow.Raised)
        self.line_vertical.setLineWidth(5)
        self.line_vertical.setFrameShape(QFrame.Shape.VLine)

        self.gridLayout.addWidget(self.line_vertical, 0, 2, 1, 1)

        self.buttonBox = QDialogButtonBox(settings_window)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel|QDialogButtonBox.StandardButton.Ok)

        self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 4)

        self.left = QWidget(settings_window)
        self.left.setObjectName(u"left")
        self.gridLayout_3 = QGridLayout(self.left)
        self.gridLayout_3.setSpacing(9)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.gridLayout_3.setContentsMargins(3, 3, 3, 3)
        self.logging = QWidget(self.left)
        self.logging.setObjectName(u"logging")
        self.gridLayout_4 = QGridLayout(self.logging)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.label_9 = QLabel(self.logging)
        self.label_9.setObjectName(u"label_9")
        font = QFont()
        font.setBold(True)
        self.label_9.setFont(font)

        self.gridLayout_4.addWidget(self.label_9, 0, 0, 1, 1)

        self.spinBox_logfile_size = QSpinBox(self.logging)
        self.spinBox_logfile_size.setObjectName(u"spinBox_logfile_size")
        self.spinBox_logfile_size.setValue(1)

        self.gridLayout_4.addWidget(self.spinBox_logfile_size, 2, 1, 1, 1)

        self.spinBox_logfile_number = QSpinBox(self.logging)
        self.spinBox_logfile_number.setObjectName(u"spinBox_logfile_number")
        self.spinBox_logfile_number.setValue(1)

        self.gridLayout_4.addWidget(self.spinBox_logfile_number, 2, 2, 1, 1)

        self.label_17 = QLabel(self.logging)
        self.label_17.setObjectName(u"label_17")

        self.gridLayout_4.addWidget(self.label_17, 1, 2, 1, 1)

        self.label_16 = QLabel(self.logging)
        self.label_16.setObjectName(u"label_16")

        self.gridLayout_4.addWidget(self.label_16, 1, 1, 1, 1)

        self.comboBox_log_level = QComboBox(self.logging)
        self.comboBox_log_level.setObjectName(u"comboBox_log_level")

        self.gridLayout_4.addWidget(self.comboBox_log_level, 2, 0, 1, 1)

        self.label_15 = QLabel(self.logging)
        self.label_15.setObjectName(u"label_15")

        self.gridLayout_4.addWidget(self.label_15, 1, 0, 1, 1)

        self.line_8 = QFrame(self.logging)
        self.line_8.setObjectName(u"line_8")
        self.line_8.setFrameShadow(QFrame.Shadow.Raised)
        self.line_8.setLineWidth(5)
        self.line_8.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_4.addWidget(self.line_8, 3, 0, 1, 3)


        self.gridLayout_3.addWidget(self.logging, 1, 0, 1, 1)

        self.saving = QWidget(self.left)
        self.saving.setObjectName(u"saving")
        self.gridLayout_2 = QGridLayout(self.saving)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.radioButton_smart_backups = QRadioButton(self.saving)
        self.radioButton_smart_backups.setObjectName(u"radioButton_smart_backups")
        self.radioButton_smart_backups.setChecked(True)

        self.gridLayout_2.addWidget(self.radioButton_smart_backups, 3, 2, 1, 1)

        self.checkBox_backup_before_run = QCheckBox(self.saving)
        self.checkBox_backup_before_run.setObjectName(u"checkBox_backup_before_run")

        self.gridLayout_2.addWidget(self.checkBox_backup_before_run, 2, 1, 1, 1)

        self.label = QLabel(self.saving)
        self.label.setObjectName(u"label")
        self.label.setMaximumSize(QSize(16777215, 17))
        font1 = QFont()
        font1.setBold(True)
        font1.setItalic(False)
        self.label.setFont(font1)

        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 3)

        self.checkBox_autosave = QCheckBox(self.saving)
        self.checkBox_autosave.setObjectName(u"checkBox_autosave")

        self.gridLayout_2.addWidget(self.checkBox_autosave, 2, 0, 1, 1)

        self.checkBox_password = QCheckBox(self.saving)
        self.checkBox_password.setObjectName(u"checkBox_password")

        self.gridLayout_2.addWidget(self.checkBox_password, 1, 0, 1, 3)

        self.checkBox_autosave_run = QCheckBox(self.saving)
        self.checkBox_autosave_run.setObjectName(u"checkBox_autosave_run")

        self.gridLayout_2.addWidget(self.checkBox_autosave_run, 2, 2, 1, 1)

        self.checkBox_new_file_each_run = QCheckBox(self.saving)
        self.checkBox_new_file_each_run.setObjectName(u"checkBox_new_file_each_run")

        self.gridLayout_2.addWidget(self.checkBox_new_file_each_run, 5, 0, 1, 3)

        self.line_4 = QFrame(self.saving)
        self.line_4.setObjectName(u"line_4")
        self.line_4.setFrameShadow(QFrame.Shadow.Raised)
        self.line_4.setLineWidth(5)
        self.line_4.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_2.addWidget(self.line_4, 7, 0, 1, 3)

        self.spinBox_backup_number = QSpinBox(self.saving)
        self.spinBox_backup_number.setObjectName(u"spinBox_backup_number")

        self.gridLayout_2.addWidget(self.spinBox_backup_number, 4, 1, 1, 1)

        self.radioButton_all_backups = QRadioButton(self.saving)
        self.radioButton_all_backups.setObjectName(u"radioButton_all_backups")

        self.gridLayout_2.addWidget(self.radioButton_all_backups, 3, 0, 1, 1)

        self.radioButton_n_backups = QRadioButton(self.saving)
        self.radioButton_n_backups.setObjectName(u"radioButton_n_backups")

        self.gridLayout_2.addWidget(self.radioButton_n_backups, 3, 1, 1, 1)

        self.label_12 = QLabel(self.saving)
        self.label_12.setObjectName(u"label_12")

        self.gridLayout_2.addWidget(self.label_12, 6, 0, 1, 1)

        self.pathButton_config_path = Path_Button_Edit(self.saving)
        self.pathButton_config_path.setObjectName(u"pathButton_config_path")

        self.gridLayout_2.addWidget(self.pathButton_config_path, 6, 1, 1, 2)


        self.gridLayout_3.addWidget(self.saving, 0, 0, 1, 1)

        self.sounds = QWidget(self.left)
        self.sounds.setObjectName(u"sounds")
        self.gridLayout_8 = QGridLayout(self.sounds)
        self.gridLayout_8.setObjectName(u"gridLayout_8")
        self.gridLayout_8.setContentsMargins(0, 0, 0, 0)
        self.checkBox_play_camel_on_error = QCheckBox(self.sounds)
        self.checkBox_play_camel_on_error.setObjectName(u"checkBox_play_camel_on_error")

        self.gridLayout_8.addWidget(self.checkBox_play_camel_on_error, 1, 0, 1, 1)

        self.label_11 = QLabel(self.sounds)
        self.label_11.setObjectName(u"label_11")
        self.label_11.setFont(font)

        self.gridLayout_8.addWidget(self.label_11, 0, 0, 1, 1)

        self.line = QFrame(self.sounds)
        self.line.setObjectName(u"line")
        self.line.setFrameShadow(QFrame.Shadow.Raised)
        self.line.setLineWidth(5)
        self.line.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_8.addWidget(self.line, 2, 0, 1, 1)


        self.gridLayout_3.addWidget(self.sounds, 3, 0, 1, 1)

        self.number_format = QWidget(self.left)
        self.number_format.setObjectName(u"number_format")
        self.gridLayout_6 = QGridLayout(self.number_format)
        self.gridLayout_6.setObjectName(u"gridLayout_6")
        self.gridLayout_6.setContentsMargins(0, 0, 0, 0)
        self.label_3 = QLabel(self.number_format)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setMaximumSize(QSize(16777215, 17))
        self.label_3.setFont(font)

        self.gridLayout_6.addWidget(self.label_3, 0, 0, 1, 3)

        self.radioButton_mixed = QRadioButton(self.number_format)
        self.radioButton_mixed.setObjectName(u"radioButton_mixed")

        self.gridLayout_6.addWidget(self.radioButton_mixed, 1, 2, 1, 1)

        self.radioButton_plain_numbers = QRadioButton(self.number_format)
        self.radioButton_plain_numbers.setObjectName(u"radioButton_plain_numbers")
        self.radioButton_plain_numbers.setCheckable(True)
        self.radioButton_plain_numbers.setChecked(True)

        self.gridLayout_6.addWidget(self.radioButton_plain_numbers, 1, 0, 1, 1)

        self.spinBox_n_decimals = QSpinBox(self.number_format)
        self.spinBox_n_decimals.setObjectName(u"spinBox_n_decimals")
        self.spinBox_n_decimals.setValue(2)

        self.gridLayout_6.addWidget(self.spinBox_n_decimals, 3, 1, 1, 2)

        self.spinBox_scientific_from = QSpinBox(self.number_format)
        self.spinBox_scientific_from.setObjectName(u"spinBox_scientific_from")
        self.spinBox_scientific_from.setValue(3)

        self.gridLayout_6.addWidget(self.spinBox_scientific_from, 2, 2, 1, 1)

        self.label_4 = QLabel(self.number_format)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setMaximumSize(QSize(16777215, 17))

        self.gridLayout_6.addWidget(self.label_4, 3, 0, 1, 1)

        self.radioButton_scientific = QRadioButton(self.number_format)
        self.radioButton_scientific.setObjectName(u"radioButton_scientific")

        self.gridLayout_6.addWidget(self.radioButton_scientific, 1, 1, 1, 1)

        self.line_7 = QFrame(self.number_format)
        self.line_7.setObjectName(u"line_7")
        self.line_7.setFrameShadow(QFrame.Shadow.Raised)
        self.line_7.setLineWidth(5)
        self.line_7.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_6.addWidget(self.line_7, 4, 0, 1, 3)


        self.gridLayout_3.addWidget(self.number_format, 5, 0, 1, 1)

        self.theme = QWidget(self.left)
        self.theme.setObjectName(u"theme")
        self.gridLayout_5 = QGridLayout(self.theme)
        self.gridLayout_5.setObjectName(u"gridLayout_5")
        self.gridLayout_5.setContentsMargins(0, 0, 0, 0)
        self.checkBox_dark = QCheckBox(self.theme)
        self.checkBox_dark.setObjectName(u"checkBox_dark")

        self.gridLayout_5.addWidget(self.checkBox_dark, 1, 2, 1, 1)

        self.label_2 = QLabel(self.theme)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setMaximumSize(QSize(16777215, 17))
        self.label_2.setFont(font)

        self.gridLayout_5.addWidget(self.label_2, 0, 0, 1, 3)

        self.comboBox_theme = QComboBox(self.theme)
        self.comboBox_theme.setObjectName(u"comboBox_theme")

        self.gridLayout_5.addWidget(self.comboBox_theme, 1, 0, 1, 1)

        self.comboBox_material_theme = QComboBox(self.theme)
        self.comboBox_material_theme.setObjectName(u"comboBox_material_theme")

        self.gridLayout_5.addWidget(self.comboBox_material_theme, 1, 1, 1, 1)

        self.line_2 = QFrame(self.theme)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShadow(QFrame.Shadow.Raised)
        self.line_2.setLineWidth(5)
        self.line_2.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_5.addWidget(self.line_2, 2, 0, 1, 3)


        self.gridLayout_3.addWidget(self.theme, 4, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_3.addItem(self.verticalSpacer, 6, 0, 1, 1)


        self.gridLayout.addWidget(self.left, 0, 0, 2, 2)

        self.right = QWidget(settings_window)
        self.right.setObjectName(u"right")
        self.gridLayout_9 = QGridLayout(self.right)
        self.gridLayout_9.setSpacing(9)
        self.gridLayout_9.setObjectName(u"gridLayout_9")
        self.gridLayout_9.setContentsMargins(3, 3, 3, 3)
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(-1, -1, -1, 0)
        self.checkBox_enable_Api = QCheckBox(self.right)
        self.checkBox_enable_Api.setObjectName(u"checkBox_enable_Api")
        self.checkBox_enable_Api.setFont(font)

        self.horizontalLayout_2.addWidget(self.checkBox_enable_Api)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_2.addItem(self.horizontalSpacer_2)

        self.label_14 = QLabel(self.right)
        self.label_14.setObjectName(u"label_14")
        self.label_14.setFont(font)

        self.horizontalLayout_2.addWidget(self.label_14)

        self.lineEdit_api_port = QLineEdit(self.right)
        self.lineEdit_api_port.setObjectName(u"lineEdit_api_port")

        self.horizontalLayout_2.addWidget(self.lineEdit_api_port)


        self.gridLayout_9.addLayout(self.horizontalLayout_2, 6, 0, 1, 1)

        self.drivers = QWidget(self.right)
        self.drivers.setObjectName(u"drivers")
        self.gridLayout_10 = QGridLayout(self.drivers)
        self.gridLayout_10.setObjectName(u"gridLayout_10")
        self.gridLayout_10.setContentsMargins(0, 0, 0, 0)
        self.label_13 = QLabel(self.drivers)
        self.label_13.setObjectName(u"label_13")
        self.label_13.setFont(font)

        self.gridLayout_10.addWidget(self.label_13, 0, 0, 1, 1)

        self.label_8 = QLabel(self.drivers)
        self.label_8.setObjectName(u"label_8")
        self.label_8.setMaximumSize(QSize(16777215, 17))

        self.gridLayout_10.addWidget(self.label_8, 1, 0, 1, 1)

        self.line_5 = QFrame(self.drivers)
        self.line_5.setObjectName(u"line_5")
        self.line_5.setFrameShadow(QFrame.Shadow.Raised)
        self.line_5.setLineWidth(5)
        self.line_5.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_10.addWidget(self.line_5, 2, 0, 1, 3)

        self.pathButton_device_path = Path_Button_Edit(self.drivers)
        self.pathButton_device_path.setObjectName(u"pathButton_device_path")

        self.gridLayout_10.addWidget(self.pathButton_device_path, 1, 1, 1, 2)


        self.gridLayout_9.addWidget(self.drivers, 1, 0, 1, 1)

        self.updates = QWidget(self.right)
        self.updates.setObjectName(u"updates")
        self.gridLayout_13 = QGridLayout(self.updates)
        self.gridLayout_13.setObjectName(u"gridLayout_13")
        self.gridLayout_13.setContentsMargins(0, 0, 0, 0)
        self.line_10 = QFrame(self.updates)
        self.line_10.setObjectName(u"line_10")
        self.line_10.setFrameShadow(QFrame.Shadow.Raised)
        self.line_10.setLineWidth(5)
        self.line_10.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_13.addWidget(self.line_10, 2, 0, 1, 1)

        self.label_23 = QLabel(self.updates)
        self.label_23.setObjectName(u"label_23")
        self.label_23.setFont(font)
        self.label_23.setStyleSheet(u"")

        self.gridLayout_13.addWidget(self.label_23, 0, 0, 1, 1)

        self.checkBox_auto_check_updates = QCheckBox(self.updates)
        self.checkBox_auto_check_updates.setObjectName(u"checkBox_auto_check_updates")

        self.gridLayout_13.addWidget(self.checkBox_auto_check_updates, 1, 0, 1, 1)


        self.gridLayout_9.addWidget(self.updates, 4, 0, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_9.addItem(self.verticalSpacer_2, 22, 0, 1, 1)

        self.label_ApiKey = QLabel(self.right)
        self.label_ApiKey.setObjectName(u"label_ApiKey")
        self.label_ApiKey.setStyleSheet(u"font-weight: bold;")

        self.gridLayout_9.addWidget(self.label_ApiKey, 5, 0, 1, 1)

        self.files = QWidget(self.right)
        self.files.setObjectName(u"files")
        self.gridLayout_7 = QGridLayout(self.files)
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.gridLayout_7.setContentsMargins(0, 0, 0, 0)
        self.label_5 = QLabel(self.files)
        self.label_5.setObjectName(u"label_5")
        self.label_5.setMaximumSize(QSize(16777215, 17))
        self.label_5.setFont(font)

        self.gridLayout_7.addWidget(self.label_5, 0, 0, 1, 1)

        self.label_7 = QLabel(self.files)
        self.label_7.setObjectName(u"label_7")
        self.label_7.setMaximumSize(QSize(16777215, 17))

        self.gridLayout_7.addWidget(self.label_7, 2, 0, 1, 1)

        self.label_10 = QLabel(self.files)
        self.label_10.setObjectName(u"label_10")

        self.gridLayout_7.addWidget(self.label_10, 3, 0, 1, 1)

        self.label_6 = QLabel(self.files)
        self.label_6.setObjectName(u"label_6")
        self.label_6.setMaximumSize(QSize(16777215, 17))

        self.gridLayout_7.addWidget(self.label_6, 1, 0, 1, 1)

        self.pathButton_py_files = Path_Button_Edit(self.files)
        self.pathButton_py_files.setObjectName(u"pathButton_py_files")

        self.gridLayout_7.addWidget(self.pathButton_py_files, 1, 1, 1, 1)

        self.line_6 = QFrame(self.files)
        self.line_6.setObjectName(u"line_6")
        self.line_6.setFrameShadow(QFrame.Shadow.Raised)
        self.line_6.setLineWidth(5)
        self.line_6.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_7.addWidget(self.line_6, 6, 0, 1, 2)

        self.pathButton_meas_files = Path_Button_Edit(self.files)
        self.pathButton_meas_files.setObjectName(u"pathButton_meas_files")

        self.gridLayout_7.addWidget(self.pathButton_meas_files, 2, 1, 1, 1)

        self.spinBox_n_databroker_files = QSpinBox(self.files)
        self.spinBox_n_databroker_files.setObjectName(u"spinBox_n_databroker_files")

        self.gridLayout_7.addWidget(self.spinBox_n_databroker_files, 5, 1, 1, 1)

        self.label_20 = QLabel(self.files)
        self.label_20.setObjectName(u"label_20")

        self.gridLayout_7.addWidget(self.label_20, 5, 0, 1, 1)

        self.lineEdit_catalog_name = QLineEdit(self.files)
        self.lineEdit_catalog_name.setObjectName(u"lineEdit_catalog_name")

        self.gridLayout_7.addWidget(self.lineEdit_catalog_name, 3, 1, 1, 1)


        self.gridLayout_9.addWidget(self.files, 0, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(-1, -1, -1, 0)
        self.pushButton_generate_Api_key = QPushButton(self.right)
        self.pushButton_generate_Api_key.setObjectName(u"pushButton_generate_Api_key")
        font2 = QFont()
        font2.setPointSize(9)
        font2.setBold(True)
        self.pushButton_generate_Api_key.setFont(font2)
        self.pushButton_generate_Api_key.setStyleSheet(u"QPushButton {\n"
"    background-color: #0db002;\n"
"}\n"
"\n"
"QPushButton:disabled {\n"
"    background-color:#4fa14a; /* A grayish-green color */\n"
"}")

        self.horizontalLayout.addWidget(self.pushButton_generate_Api_key)

        self.pushButton_delete_Api_keys = QPushButton(self.right)
        self.pushButton_delete_Api_keys.setObjectName(u"pushButton_delete_Api_keys")
        self.pushButton_delete_Api_keys.setFont(font2)
        self.pushButton_delete_Api_keys.setStyleSheet(u"QPushButton {\n"
"    background-color: #fa0000;\n"
"}\n"
"\n"
"QPushButton:disabled {\n"
"    background-color: #A52A2A; /* A grayish-red color */\n"
"}")

        self.horizontalLayout.addWidget(self.pushButton_delete_Api_keys)


        self.gridLayout_9.addLayout(self.horizontalLayout, 11, 0, 4, 1)

        self.extensions = QWidget(self.right)
        self.extensions.setObjectName(u"extensions")
        self.gridLayout_12 = QGridLayout(self.extensions)
        self.gridLayout_12.setObjectName(u"gridLayout_12")
        self.gridLayout_12.setContentsMargins(0, 0, 0, 0)
        self.label_21 = QLabel(self.extensions)
        self.label_21.setObjectName(u"label_21")
        self.label_21.setMaximumSize(QSize(16777215, 17))

        self.gridLayout_12.addWidget(self.label_21, 1, 0, 1, 1)

        self.label_22 = QLabel(self.extensions)
        self.label_22.setObjectName(u"label_22")
        self.label_22.setFont(font)

        self.gridLayout_12.addWidget(self.label_22, 0, 0, 1, 2)

        self.pathButton_extension_path = Path_Button_Edit(self.extensions)
        self.pathButton_extension_path.setObjectName(u"pathButton_extension_path")

        self.gridLayout_12.addWidget(self.pathButton_extension_path, 1, 1, 1, 1)

        self.line_9 = QFrame(self.extensions)
        self.line_9.setObjectName(u"line_9")
        self.line_9.setFrameShadow(QFrame.Shadow.Raised)
        self.line_9.setLineWidth(5)
        self.line_9.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_12.addWidget(self.line_9, 2, 0, 1, 2)


        self.gridLayout_9.addWidget(self.extensions, 3, 0, 1, 1)

        self.eln = QWidget(self.right)
        self.eln.setObjectName(u"eln")
        self.gridLayout_11 = QGridLayout(self.eln)
        self.gridLayout_11.setObjectName(u"gridLayout_11")
        self.gridLayout_11.setContentsMargins(0, 0, 0, 0)
        self.label_19 = QLabel(self.eln)
        self.label_19.setObjectName(u"label_19")

        self.gridLayout_11.addWidget(self.label_19, 1, 0, 1, 1)

        self.lineEdit_oasis = QLineEdit(self.eln)
        self.lineEdit_oasis.setObjectName(u"lineEdit_oasis")

        self.gridLayout_11.addWidget(self.lineEdit_oasis, 1, 1, 1, 1)

        self.label_18 = QLabel(self.eln)
        self.label_18.setObjectName(u"label_18")
        self.label_18.setFont(font)

        self.gridLayout_11.addWidget(self.label_18, 0, 0, 1, 2)

        self.line_3 = QFrame(self.eln)
        self.line_3.setObjectName(u"line_3")
        self.line_3.setFrameShadow(QFrame.Shadow.Raised)
        self.line_3.setLineWidth(5)
        self.line_3.setFrameShape(QFrame.Shape.HLine)

        self.gridLayout_11.addWidget(self.line_3, 2, 0, 1, 2)


        self.gridLayout_9.addWidget(self.eln, 2, 0, 1, 1)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.Api_key_lineEdit = QLineEdit(self.right)
        self.Api_key_lineEdit.setObjectName(u"Api_key_lineEdit")
        self.Api_key_lineEdit.setReadOnly(True)

        self.horizontalLayout_4.addWidget(self.Api_key_lineEdit)

        self.pushButton_copy_Api_key_clipboard = QPushButton(self.right)
        self.pushButton_copy_Api_key_clipboard.setObjectName(u"pushButton_copy_Api_key_clipboard")
        self.pushButton_copy_Api_key_clipboard.setEnabled(False)
        self.pushButton_copy_Api_key_clipboard.setStyleSheet(u"")
        icon = QIcon()
        icon.addFile(u":/images/graphics/copy_to_clipboard.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.pushButton_copy_Api_key_clipboard.setIcon(icon)
        self.pushButton_copy_Api_key_clipboard.setIconSize(QSize(16, 16))

        self.horizontalLayout_4.addWidget(self.pushButton_copy_Api_key_clipboard)


        self.gridLayout_9.addLayout(self.horizontalLayout_4, 15, 0, 1, 1)

        self.pushButton_API_docu = QPushButton(self.right)
        self.pushButton_API_docu.setObjectName(u"pushButton_API_docu")

        self.gridLayout_9.addWidget(self.pushButton_API_docu, 7, 0, 1, 1)


        self.gridLayout.addWidget(self.right, 0, 3, 1, 1)


        self.retranslateUi(settings_window)
        self.buttonBox.accepted.connect(settings_window.accept)
        self.buttonBox.rejected.connect(settings_window.reject)

        QMetaObject.connectSlotsByName(settings_window)
    # setupUi

    def retranslateUi(self, settings_window):
        settings_window.setWindowTitle(QCoreApplication.translate("settings_window", u"Dialog", None))
        self.label_9.setText(QCoreApplication.translate("settings_window", u"Logging", None))
        self.label_17.setText(QCoreApplication.translate("settings_window", u"old logfile backups", None))
        self.label_16.setText(QCoreApplication.translate("settings_window", u"max. logfile size (MB)", None))
        self.label_15.setText(QCoreApplication.translate("settings_window", u"Log-Level", None))
        self.radioButton_smart_backups.setText(QCoreApplication.translate("settings_window", u"\"smart\" backups", None))
        self.checkBox_backup_before_run.setText(QCoreApplication.translate("settings_window", u"backup before run", None))
        self.label.setText(QCoreApplication.translate("settings_window", u"Saving", None))
        self.checkBox_autosave.setText(QCoreApplication.translate("settings_window", u"autosave on closing", None))
        self.checkBox_password.setText(QCoreApplication.translate("settings_window", u"password protection", None))
        self.checkBox_autosave_run.setText(QCoreApplication.translate("settings_window", u"autosave before run", None))
        self.checkBox_new_file_each_run.setText(QCoreApplication.translate("settings_window", u"save each measurement run in new file", None))
        self.radioButton_all_backups.setText(QCoreApplication.translate("settings_window", u"keep all backups", None))
        self.radioButton_n_backups.setText(QCoreApplication.translate("settings_window", u"keep only...", None))
        self.label_12.setText(QCoreApplication.translate("settings_window", u"path for configuration files:", None))
        self.checkBox_play_camel_on_error.setText(QCoreApplication.translate("settings_window", u"Play Camel-Roar on error", None))
        self.label_11.setText(QCoreApplication.translate("settings_window", u"Sounds", None))
        self.label_3.setText(QCoreApplication.translate("settings_window", u"Number-Formatting (only visual)", None))
        self.radioButton_mixed.setText(QCoreApplication.translate("settings_window", u"Scientific from 1e...", None))
        self.radioButton_plain_numbers.setText(QCoreApplication.translate("settings_window", u"Plain", None))
        self.label_4.setText(QCoreApplication.translate("settings_window", u"# decimals:", None))
        self.radioButton_scientific.setText(QCoreApplication.translate("settings_window", u"Scientific", None))
        self.checkBox_dark.setText(QCoreApplication.translate("settings_window", u"dark mode", None))
        self.label_2.setText(QCoreApplication.translate("settings_window", u"Theme", None))
        self.checkBox_enable_Api.setText(QCoreApplication.translate("settings_window", u"Enable API", None))
        self.label_14.setText(QCoreApplication.translate("settings_window", u"API Port:", None))
        self.lineEdit_api_port.setPlaceholderText(QCoreApplication.translate("settings_window", u"API Port", None))
        self.label_13.setText(QCoreApplication.translate("settings_window", u"Drivers", None))
        self.label_8.setText(QCoreApplication.translate("settings_window", u"Local drivers path", None))
        self.label_23.setText(QCoreApplication.translate("settings_window", u"Updates", None))
        self.checkBox_auto_check_updates.setText(QCoreApplication.translate("settings_window", u"automatically search for updates", None))
        self.label_ApiKey.setText(QCoreApplication.translate("settings_window", u"API Key", None))
        self.label_5.setText(QCoreApplication.translate("settings_window", u"Files", None))
        self.label_7.setText(QCoreApplication.translate("settings_window", u"Measurement-Data Path", None))
        self.label_10.setText(QCoreApplication.translate("settings_window", u"Databroker catalog-name", None))
        self.label_6.setText(QCoreApplication.translate("settings_window", u"Python-Files Path", None))
#if QT_CONFIG(tooltip)
        self.spinBox_n_databroker_files.setToolTip(QCoreApplication.translate("settings_window", u"with \"0\" all files will be kept", None))
#endif // QT_CONFIG(tooltip)
        self.label_20.setText(QCoreApplication.translate("settings_window", u"# databroker files", None))
        self.pushButton_generate_Api_key.setText(QCoreApplication.translate("settings_window", u"Generate\n"
"API Key", None))
        self.pushButton_delete_Api_keys.setText(QCoreApplication.translate("settings_window", u"Delete ALL\n"
"API keys", None))
        self.label_21.setText(QCoreApplication.translate("settings_window", u"Local extensions path", None))
        self.label_22.setText(QCoreApplication.translate("settings_window", u"Extensions", None))
        self.label_19.setText(QCoreApplication.translate("settings_window", u"NOMAD Oasis URL", None))
        self.label_18.setText(QCoreApplication.translate("settings_window", u"ELN integration", None))
        self.pushButton_copy_Api_key_clipboard.setText(QCoreApplication.translate("settings_window", u"Copy key to clipboard", None))
        self.pushButton_API_docu.setText(QCoreApplication.translate("settings_window", u"API documentation", None))
    # retranslateUi




import requests
import re
from requests.auth import HTTPBasicAuth


def execute_camels_api_call(
    host: str,
    port: str,
    api_type: str,
    message_body: dict,
    authentication_type: str,
    authentication_string: str,
    selected_camels_function_index: int,
    camels_function_parameters: dict,
):
    """
    This function executes the CAMELS API call and returns the results.
    """
    # Make sure there is HTTP schema in the host
    if not host.startswith("http"):
        host = f"http://{host}"
    # Use port 80 for http and 443 for https if no port is provided
    if port == "":
        if host.startswith("https"):
            port = "443"
        else:
            port = "80"
    if api_type != "CAMELS":
        raise ValueError(
            "The API type is not CAMELS. This function is only for executing CAMELS API functions."
        )
    api_functions = get_available_camels_api_functions(
        host, port, camels_function_parameters
    )
    # Get the selected function
    selected_function = api_functions[selected_camels_function_index]
    # Check the HTTP method
    if selected_function["method"] == "GET":
        # Check the authentication type
        if authentication_type == "Bearer Token":
            result = requests.get(
                f"{host}:{port}{selected_function['formatted_path']}",
                headers={"Authorization": f"Bearer {authentication_string}"},
            )
        elif authentication_type == "HTTP Basic":
            raise ValueError(
                "HTTP Basic authentication is not supported for CAMELS requests."
            )
        elif authentication_type == "None":
            raise ValueError("No authentication is not supported for CAMELS requests.")

    elif selected_function["method"] == "POST":
        # Check the authentication type
        if authentication_type == "Bearer Token":
            result = requests.post(
                f"{host}:{port}{selected_function['formatted_path']}",
                headers={"Authorization": f"Bearer {authentication_string}"},
                json=message_body,
            )
        elif authentication_type == "HTTP Basic":
            raise ValueError(
                "HTTP Basic authentication is not supported for CAMELS requests."
            )
        elif authentication_type == "None":
            raise ValueError("No authentication is not supported for CAMELS requests.")

    return result.json()


def execute_generic_api_call(
    host: str,
    port: str,
    api_type: str,
    api_url: str,
    http_method: str,
    message_body: dict,
    authentication_type: str,
    authentication_string: str,
    http_parameters: dict,

):
    """
    This function executes the generic API call and returns the results.
    """
    # Make sure there is HTTP schema in the host
    if not host.startswith("http"):
        host = f"http://{host}"
    # Use port 80 for http and 443 for https if no port is provided
    if port == "":
        if host.startswith("https"):
            port = "443"
        else:
            port = "80"
    if api_type != "Generic":
        raise ValueError(
            "The API type is not Generic. This function is only for executing Generic API functions."
        )
    
    # Add the http_parameters to the api_url
    if http_parameters:
        print(http_parameters)
        api_url = api_url + "?"
        parameters = http_parameters['Parameter']
        values = http_parameters['Value']
        for param, value in zip(parameters, values):
            api_url = api_url + f"{param}={value}&"
        api_url = api_url[:-1]
        print(api_url)
        print('_'   * 50)
    
    
    # Check the HTTP method
    if http_method == "GET":
        # Check the authentication type
        if authentication_type == "Bearer Token":
            result = requests.get(
                f"{host}:{port}{api_url}",
                headers={"Authorization": f"Bearer {authentication_string}"},
            )
        elif authentication_type == "HTTP Basic":
            result = requests.get(
                f"{host}:{port}{api_url}",
                auth=HTTPBasicAuth(*authentication_string.split(":")),
            )
        elif authentication_type == "None":
            result = requests.get(f"{host}:{port}{api_url}")

    elif http_method == "POST":
        # Check the authentication type
        if authentication_type == "Bearer Token":
            result = requests.post(
                f"{host}:{port}{api_url}",
                headers={"Authorization": f"Bearer {authentication_string}"},
                json=message_body,
            )
        elif authentication_type == "HTTP Basic":
            result = requests.post(
                f"{host}:{port}{api_url}",
                auth=HTTPBasicAuth(*authentication_string.split(":")),
                json=message_body,
            )
        elif authentication_type == "None":
            result = requests.post(
                f"{host}:{port}{api_url}",
                json=message_body,
            )
    elif http_method == "DELETE":
        # Check the authentication type
        if authentication_type == "Bearer Token":
            result = requests.delete(
                f"{host}:{port}{api_url}",
                headers={"Authorization": f"Bearer {authentication_string}"},
            )
        elif authentication_type == "HTTP Basic":
            result = requests.delete(
                f"{host}:{port}{api_url}",
                auth=HTTPBasicAuth(*authentication_string.split(":")),
            )
        elif authentication_type == "None":
            result = requests.delete(f"{host}:{port}{api_url}")

        elif http_method == "PATCH":
            # Check the authentication type
            if authentication_type == "Bearer Token":
                result = requests.patch(
                    f"{host}:{port}{api_url}",
                    headers={"Authorization": f"Bearer {authentication_string}"},
                    json=message_body,
                )
            elif authentication_type == "HTTP Basic":
                result = requests.patch(
                    f"{host}:{port}{api_url}",
                    auth=HTTPBasicAuth(*authentication_string.split(":")),
                    json=message_body,
                )
            elif authentication_type == "None":
                result = requests.patch(
                    f"{host}:{port}{api_url}",
                    json=message_body,
                )

            elif http_method == "PUT":
                # Check the authentication type
                if authentication_type == "Bearer Token":
                    result = requests.put(
                        f"{host}:{port}{api_url}",
                        headers={"Authorization": f"Bearer {authentication_string}"},
                        json=message_body,
                    )
                elif authentication_type == "HTTP Basic":
                    result = requests.put(
                        f"{host}:{port}{api_url}",
                        auth=HTTPBasicAuth(*authentication_string.split(":")),
                        json=message_body,
                    )
                elif authentication_type == "None":
                    result = requests.put(
                        f"{host}:{port}{api_url}",
                        json=message_body,
                    )
    if result.status_code == 403:
        raise ValueError("The authentication failed. Please check your credentials.")
    return result.json()


def save_API_response_to_variable(response, namespace, protocol_step_name):
    """
    This function saves the API response to the namespace. This allows you to use the response values in the script and following protocol steps.
    """
    # For key value pair in response dictionary save it to the namesapce
    for key, value in response.items():
        namespace[f"{protocol_step_name}_{key}"] = value
    return None


def evaluate_message_body(message_body, eva):
    """
    This function evaluates the values of the message_body items if the value is empty.
    This allows you to use variables and values that are obtained during the script execution.
    """
    # For key value pair in message_body dictionary evaluate the value
    for key, value in message_body["variables"].items():
        if value == "":
            message_body["variables"][key] = eva.eval(key)
    return message_body


def get_available_camels_api_functions(host, port, camels_function_parameters):
    """
    This function gets all the available CAMELS API functions.
    """
    # Get all the available functions from the API
    result = requests.get(f"{host}:{port}/openapi.json")
    # Parse the JSON response
    openapi_schema = result.json()
    paths = openapi_schema.get("paths", {})
    # Create the formatted path
    # Extract the function names and descriptions
    api_functions = []
    for path, methods in paths.items():
        # Extract variables from the path
        path_variables = re.findall(r"\{(\w+)\}", path)

        # Check if all variables are in the parameters dictionary
        if all(var in camels_function_parameters for var in path_variables):
            formatted_path = path.format(**camels_function_parameters)
        else:
            formatted_path = "Not modified"  # or handle it accordingly
        for method, details in methods.items():
            summary = details.get("summary", "No summary available")
            operation_id = details.get("operationId", "Unnamed function")
            description = details.get("description", "No description available")
            parameters = details.get("parameters", [])
            api_functions.append(
                {
                    "summary": summary,
                    "method": method.upper(),
                    "path": path,
                    "operation_id": operation_id,
                    "description": description,
                    "parameters": parameters,
                    "formatted_path": formatted_path,
                }
            )
    return api_functions



from PySide6.QtCore import QItemSelectionModel
from PySide6.QtGui import QUndoCommand


class CommandMoveStep(QUndoCommand):
    """
    Using a command for moving a step inside the protocol treeView to be able to
    reverse the action.

    Parameters
    ----------
    treeView : QTreeView
        The TreeView displaying the protocol sequence
    item_model : QStandardItemModel
        The treeview's item model
    up_down : int
        The direction in which to move the step, if positive it moves down, if
        negative it moves up.
    in_out : int
        The step is moved into the step above it, if positive, if negative it is
        moved out one layer.
    loop_step_dict : dict
        The dictionary of the loopsteps, used to find the parent steps
    update_func : function
        This function is called, when the command is finished, usually used to
        update the order of the steps in the protocol data.
    """

    def __init__(
        self, treeView, item_model, up_down, in_out, loop_step_dict, update_func
    ):
        super().__init__()
        self.treeView_protocol_sequence = treeView
        self.item_model_sequence = item_model
        self.up_down = up_down
        self.in_out = in_out
        self.loop_step_dict = loop_step_dict
        self.ind = self.treeView_protocol_sequence.selectedIndexes()[0]
        self.item = self.item_model_sequence.itemFromIndex(self.ind)
        self.update_func = update_func

    def redo(self):
        """Performs moving the step."""
        parent = self.item.parent()
        if parent is None:
            if (
                self.up_down != 0
                and (self.ind.row() > 0 or self.up_down > 0)
                and (
                    self.ind.row() < self.item_model_sequence.rowCount() - 1
                    or self.up_down < 0
                )
            ):
                row = self.item_model_sequence.takeRow(self.ind.row())
                self.item_model_sequence.insertRow(self.ind.row() + self.up_down, row)
            elif self.in_out > 0 and self.ind.row() > 0:
                above = self.item_model_sequence.item(self.ind.row() - 1, 0)
                step = self.loop_step_dict[above.data()]
                if step.step_type == "If":
                    above = above.child(above.rowCount() - 1, 0)
                if self.loop_step_dict[above.data()].has_children:
                    row = self.item_model_sequence.takeRow(self.ind.row())
                    above.insertRow(above.rowCount(), row)
        else:
            if (
                self.up_down != 0
                and (self.ind.row() > 0 or self.up_down > 0)
                and (self.ind.row() < parent.rowCount() - 1 or self.up_down < 0)
            ):
                row = parent.takeRow(self.ind.row())
                parent.insertRow(self.ind.row() + self.up_down, row)
            elif self.in_out > 0 and self.ind.row() > 0:
                above = parent.child(self.ind.row() - 1, 0)
                if self.loop_step_dict[above.data()].has_children:
                    row = parent.takeRow(self.ind.row())
                    above.insertRow(above.rowCount(), row)
            elif self.in_out < 0:
                grandparent = parent.parent()
                if grandparent is None:
                    grandparent = self.item_model_sequence
                parent_row = parent.index().row()
                row = parent.takeRow(self.ind.row())
                grandparent.insertRow(parent_row + 1, row)
        self.treeView_protocol_sequence.clearSelection()
        new_ind = self.item_model_sequence.indexFromItem(self.item)
        self.treeView_protocol_sequence.selectionModel().select(
            new_ind, QItemSelectionModel.Select
        )
        self.update_func()

    def undo(self):
        """Moves the step back to its original position."""
        parent = self.item.parent()
        up_down = -self.up_down
        in_out = -self.in_out
        ind = self.item_model_sequence.indexFromItem(self.item)
        if parent is None:
            if (
                up_down != 0
                and (ind.row() > 0 or up_down > 0)
                and (ind.row() < self.item_model_sequence.rowCount() - 1 or up_down < 0)
            ):
                row = self.item_model_sequence.takeRow(ind.row())
                self.item_model_sequence.insertRow(ind.row() + up_down, row)
            elif in_out > 0 and ind.row() > 0:
                above = self.item_model_sequence.item(ind.row() - 1, 0)
                if self.loop_step_dict[above.data()].has_children:
                    row = self.item_model_sequence.takeRow(ind.row())
                    above.insertRow(above.rowCount(), row)
        else:
            if (
                up_down != 0
                and (ind.row() > 0 or up_down > 0)
                and (ind.row() < parent.rowCount() - 1 or up_down < 0)
            ):
                row = parent.takeRow(ind.row())
                parent.insertRow(ind.row() + up_down, row)
            elif in_out > 0 and ind.row() > 0:
                above = parent.child(ind.row() - 1, 0)
                if self.loop_step_dict[above.data()].has_children:
                    row = parent.takeRow(ind.row())
                    above.insertRow(above.rowCount(), row)
            elif in_out < 0:
                grandparent = parent.parent()
                if grandparent is None:
                    grandparent = self.item_model_sequence
                parent_row = parent.index().row()
                row = parent.takeRow(ind.row())
                grandparent.insertRow(parent_row + 1, row)
        self.treeView_protocol_sequence.clearSelection()
        new_ind = self.item_model_sequence.indexFromItem(self.item)
        self.treeView_protocol_sequence.selectionModel().select(
            new_ind, QItemSelectionModel.Select
        )
        self.update_func()



from bluesky.callbacks.core import CallbackBase
import numpy as np
import ast
import time
import scipy.constants as const

starttime = time.time()

base_namespace = {"numpy": np, "np": np, "time": 0, "const": const}
base_namespace.update({name: getattr(np, name) for name in np.__all__})
base_namespace.update({"StartTime": starttime, "ElapsedTime": time.time() - starttime})


class Evaluator(CallbackBase):
    """
    This class can be used to evaluate strings with names known in bluesky at
    any point, as well inside the run itself.
    """

    def __init__(self, *args, namespace=None, **kwargs):
        """
        Parameters
        ----------
        namespace : dict
            The namespace of the run that is added to the default base_namespace
        """
        super().__init__(*args, **kwargs)
        self.add_namespace = {}
        if namespace is not None:
            self.add_namespace = namespace
        self.namespace = dict(base_namespace)
        self.update_namespace()
        self.last_update = 0
        self.raised_exceptions = []

    def eval_string(self, eval_str: str):
        """
        Evaluates the given string through the provided namespace to another
        string. The string is split at quotation marks and the parts of the
        string are tried to be evaluated. If that is not possible, they are just
        added back to the string as is.
        Checks if the namespace is up to date before evaluating.

        Parameters
        ----------
        eval_str : str
            The string to be evaluated.


        Returns
        -------
        return_str
            The evaluated string.

        Raises
        ------
        ValueError
            If there is a bigger problem.

        """
        self.update_namespace()
        try:
            return_str = ""
            for part in eval_str.split('"' if '"' in eval_str else "'"):
                try:
                    return_str += str(eval(part, self.namespace))
                except:
                    return_str += part
            return return_str
        except Exception as err:
            raise ValueError(f"Could not evaluate {eval_str!r}") from err

    def eval(self, eval_str: str, do_not_reraise=False):
        """
        Evaluates the string within the given namespace.
        Checks if the namespace is up to date before evaluating.

        Parameters
        ----------
        eval_str : str
            The string to be evaluated.


        Returns
        -------
        Value
            The evaluated value.

        """
        self.update_namespace()
        if do_not_reraise:
            return get_eval(eval_str, self.namespace, self)
        return get_eval(eval_str, self.namespace)

    def start(self, doc):
        """
        Overwrites the `start` method to add the variable 'StartTime' to the
        namespace.
        See bluesky's documentation for more information.
        """
        self.namespace["StartTime"] = time.time()
        super().start(doc)

    def event(self, doc):
        """
        Overwrites the `event` method to add the variable 'time' to the
        namespace and update the `last_update` attribute for checking before
        evaluation.
        See bluesky's documentation for more information.
        """
        self.namespace["time"] = doc["time"]
        self.namespace.update(doc["data"])
        # self.update_namespace()
        self.last_update = doc["time"]

    def update_namespace(self):
        """
        Updates the namespace with the `add_namespace` dictionary. Since this is
        exactly the given namespace from __init__, it can be easily updated from
        outside the class.
        Also the variable 'ElapsedTime' is added to the namespace.
        """
        self.namespace.update(self.add_namespace)
        self.namespace["ElapsedTime"] = time.time() - self.namespace["StartTime"]

    def is_to_date(self, t):
        """
        Compares whether the given `t` is the same as the Evaluator's
        last_update.
        """
        return self.last_update == t


def get_eval(eval_str, namespace, evaluator=None):
    """This evaluates the given string within the given namespace.
    Most functionality is taken from bluesky.utils.call_or_eval_one."""
    # If it is a key in our namespace, look it up.
    if not isinstance(eval_str, str):
        return eval_str
    eval_str = eval_str.strip()
    try:
        # This handles field or stream names that are not valid
        # Python identifiers (e.g. ones with spaces in them).
        return namespace[eval_str]
    except KeyError:
        pass
    # Check whether it is valid Python syntax.
    try:
        ast.parse(eval_str)
    except SyntaxError as err:
        error_text = f"Could not find {eval_str!r} in namespace or evaluate it."
        if evaluator and error_text in evaluator.raised_exceptions:
            return np.nan
        if evaluator is not None:
            evaluator.raised_exceptions.append(error_text)
        raise ValueError(
            f"Could not find {eval_str!r} in namespace or parse it as a Python expression."
        ) from err
    # Try to evaluate it as a Python expression in the namespace.
    try:
        return eval(eval_str, namespace)
    except Exception as err:
        error_text = f"Could not find {eval_str!r} in namespace or evaluate it."
        if evaluator is not None:
            if evaluator and error_text in evaluator.raised_exceptions:
                return np.nan
            evaluator.raised_exceptions.append(error_text)
        raise ValueError(error_text) from err



from PySide6.QtWidgets import (
    QGridLayout,
    QDialog,
    QDialogButtonBox,
    QLabel,
    QLineEdit,
    QMessageBox,
    QWidget,
)
from PySide6.QtGui import QIcon, QCloseEvent
from PySide6.QtCore import Signal, Qt

from nomad_camels.utility import device_handling, variables_handling

from importlib import resources
from nomad_camels import graphics


class Manual_Control(QWidget):
    """
    Parent class for manual controls.

    This class provides the core functionality of a manual control in CAMELS.
    The parameters `device`, `device_list` and `ophyd_device` may or may not be
    used by the child class.

    Attributes
    ----------
    name : str
        The name of the manual control. It also determines the window title.
    device : str, None
        The device the manual control is using
    device_list : list
        A list of the devices, used. The specific implementation lies in the
        child classes.
    ophyd_device : ophyd.Device
        The device's representation in ophyd.
    """

    closing = Signal()

    def __init__(self, parent=None, title="Manual Control", control_data=None):
        super().__init__()
        # layout = QGridLayout()
        # self.setLayout(layout)
        self.control_data = control_data or {}

        self.setWindowTitle(f"{title} - NOMAD CAMELS")
        self.setWindowIcon(QIcon(str(resources.files(graphics) / "camels_icon.png")))
        self.name = title
        self.device = None
        self.ophyd_device = None
        self.instantiate_devices_thread = None
        self.device_list = []
        self.show()

    # def close(self) -> bool:
    #     # device_handling.close_devices(self.device_list)
    #     self.closing.emit()
    #     return super().close()

    def closeEvent(self, a0: QCloseEvent) -> None:
        """Overwritten, so that `self.closing` is emitted, telling the main UI
        window that this manual control is no longer opened."""
        device_handling.close_devices(self.device_list)
        self.closing.emit()
        return super().closeEvent(a0)

    def propagate_exception(self, exception):
        """Propagates an exception to the main UI window.

        Parameters
        ----------
        exception : Exception
            The exception to be propagated.
        """
        raise exception

    def start_device(self, device_name):
        """Starts a device by using the
        `device_handling.InstantiateDevicesThread` class.

        Parameters
        ----------
        device_name : str
            The name of the device to be started.
        """
        self.device = variables_handling.devices[device_name]
        if self.device:
            self.device_list = self.device.get_necessary_devices()
            self.device_list = list(set(self.device_list))
            if self.device.name in self.device_list:
                self.device_list.remove(self.device.custom_name)
            self.device_list.append(self.device.custom_name)
            self.instantiate_devices_thread = device_handling.InstantiateDevicesThread(
                self.device_list
            )
            self.instantiate_devices_thread.successful.connect(self.device_ready)
            self.setCursor(Qt.WaitCursor)
            self.setEnabled(False)
            self.instantiate_devices_thread.exception_raised.connect(
                self.propagate_exception
            )
            self.instantiate_devices_thread.start()

    def start_multiple_devices(self, device_names, channels=False):
        """Starts multiple devices at once.

        Parameters
        ----------
        device_names : list
            A list of the names of the devices to be started, or the names of the channels, if `channels` is True.
        channels : bool
            Whether 'device_names' are channel names or device names.
        """
        self.instantiate_devices_thread = device_handling.InstantiateDevicesThread(
            device_names, channels
        )
        self.instantiate_devices_thread.successful.connect(self.device_ready)
        self.setEnabled(False)
        self.instantiate_devices_thread.exception_raised.connect(
            self.propagate_exception
        )
        self.instantiate_devices_thread.start()

    def device_ready(self):
        """Called when the devices are ready to be used, i.e. when the
        `instantiate_devices_thread` is finished."""
        self.device_list = self.instantiate_devices_thread.devices
        if self.device:
            self.ophyd_device = self.device_list[self.device.custom_name]
        self.setCursor(Qt.ArrowCursor)
        self.setEnabled(True)


class Manual_Control_Config(QDialog):
    """ """

    def __init__(
        self,
        parent=None,
        control_data=None,
        title="Manual Control Config",
        control_type="",
    ):
        super().__init__(parent=parent)
        self.setWindowFlags(self.windowFlags() | Qt.WindowMaximizeButtonHint)
        layout = QGridLayout()
        self.setLayout(layout)
        self.control_type = control_type or "Manual_Control"

        self.setWindowTitle(f"{title} - NOMAD CAMELS")

        self.buttonBox = QDialogButtonBox()
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.rejected.connect(self.reject)
        self.buttonBox.accepted.connect(self.accept)

        name_label = QLabel("Name:")
        self.control_data = control_data or {}
        if "name" in self.control_data:
            name = self.control_data["name"]
        else:
            name = self.control_type
        self.lineEdit_name = QLineEdit(name)

        layout.addWidget(name_label, 0, 0)
        layout.addWidget(self.lineEdit_name, 0, 1)
        layout.addWidget(self.buttonBox, 11, 0, 1, 2)

    def accept(self):
        """ """
        self.control_data["name"] = self.lineEdit_name.text()
        self.control_data["control_type"] = self.control_type
        super().accept()

    def closeEvent(self, a0: QCloseEvent) -> None:
        """

        Parameters
        ----------
        a0: QCloseEvent :


        Returns
        -------

        """
        discard_dialog = QMessageBox.question(
            self,
            "Discard Changes?",
            f"All changes will be lost!",
            QMessageBox.Yes | QMessageBox.No,
        )
        if discard_dialog != QMessageBox.Yes:
            a0.ignore()
            return
        super().closeEvent(a0)



from PySide6.QtWidgets import (
    QRadioButton,
    QButtonGroup,
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QFrame,
    QLabel,
)
from nomad_camels.ui_widgets.add_remove_table import AddRemoveTable
from nomad_camels.main_classes.loop_step import Loop_Step, Loop_Step_Config
from nomad_camels.ui_widgets.path_button_edit import Path_Button_Edit
import os
from nomad_camels.utility import variables_handling


class Execute_Python_File(Loop_Step):
    """
    A loop step to execute a python file. The python.exe path as well as the python file path must be given.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Execute Python File"
        if step_info is None:
            step_info = {}
        # Get the python file path and the python exe path from the step info
        self.file_path = step_info["file_path"] if "file_path" in step_info else ""
        self.python_exe_path = (
            step_info["python_exe_path"] if "python_exe_path" in step_info else ""
        )
        # Get the python packages and versions from the step info
        self.python_packages_versions = (
            step_info["python_packages_versions"]
            if "python_packages_versions" in step_info
            else []
        )
        # Get the radio button selected from the step info
        self.radio_button_selected = (
            step_info["radio_button_selected"]
            if "radio_button_selected" in step_info
            else None
        )

        # Get the python package and version variables and their value from the step info
        self.variables_passing = (
            step_info["variables_passing"] if "variables_passing" in step_info else {}
        )
        # Get the variables returned by the file from the step info
        self.returned_values_variables = (
            step_info["returned_values_variables"]
            if "returned_values_variables" in step_info
            else {}
        )

        self.use_existing_env = (
            step_info["use_existing_env"] if "use_existing_env" in step_info else False
        )
        self.use_specific_packages = (
            step_info["use_specific_packages"]
            if "use_specific_packages" in step_info
            else False
        )
        self.use_camels_python = (
            step_info["use_camels_python"]
            if "use_camels_python" in step_info
            else False
        )

    def get_protocol_short_string(self, n_tabs=0):
        short_string = super().get_protocol_short_string(n_tabs)
        short_string = f"{short_string[:-1]}: {os.path.basename(self.file_path)}"
        return short_string

    def get_protocol_string(self, n_tabs=1):
        """
        This function runs the python file with the python exe using the subprocess module.
        The the cwd of the subprocess is changed to the python files location.
        """
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        # use subprocess to run the python file
        if self.variables_passing:
            formatted_strings = []
            variable_names = self.variables_passing["Variable Name"]
            values = self.variables_passing["Value"]
            for i in range(len(variable_names)):
                # Format each "name = value" pair
                if values[i] == "":
                    formatted_string = (
                        f"f'{variable_names[i]}={{eva.eval(\"{variable_names[i]}\")}}',"
                    )
                else:
                    formatted_string = f"f'{variable_names[i]}={{{values[i]}}}',"
                # Append to the list
                formatted_strings.append(formatted_string)
            # Step 4: Join all formatted strings with a space
            variables_string = " ".join(formatted_strings)
            variables_string = variables_string[:-1]
        else:
            variables_string = ""
        if self.use_existing_env:
            protocol_string += (
                f"{tabs}result_python_file = subprocess.run(['"
                f'r"{os.path.abspath(self.python_exe_path)}", '
                f'r"{os.path.abspath(self.file_path)}", '
                f"{variables_string}], "
                f'cwd=r"{os.path.abspath(os.path.dirname(self.file_path))}", '
                f"capture_output=True, text=True)\n"
                f"{tabs}helper_functions.evaluate_python_file_output(result_python_file.stdout, namespace)\n"
            )

        if self.use_specific_packages:
            protocol_string += (
                f"{tabs}result_python_file = helper_functions.create_venv_run_file_delete_venv({self.python_packages_versions}, "
                f'r"{os.path.abspath(self.file_path)}", '
                f"{variables_string})\n"
                f"{tabs}helper_functions.evaluate_python_file_output(result_python_file.stdout, namespace)\n"
            )

        if self.use_camels_python:
            import sys

            protocol_string += (
                f"{tabs}result_python_file = subprocess.run(["
                f'r"{os.path.abspath(sys.executable)}", '
                f'r"{os.path.abspath(self.file_path)}", '
                f"{variables_string}], "
                f'cwd=r"{os.path.abspath(os.path.dirname(self.file_path))}", '
                f"capture_output=True, text=True)\n"
                f"{tabs}helper_functions.evaluate_python_file_output(result_python_file.stdout, namespace)\n"
            )
        return protocol_string

    def update_variables(self):
        """ """
        # Update the global variables to have access to the variables defined to pass to the python file
        variable_names = self.variables_passing.get("Variable Name", [])
        values = self.variables_passing.get("Value", [])
        # Create a dictionary from the lists
        new_variables = dict(zip(variable_names, values))
        # remove keys and value if the key is already in variables_handling.channels.keys()
        # remove keys and value if the value is empty so that the value is taken from the running script
        for key in list(new_variables.keys()):
            if (key in variables_handling.channels.keys()) or (
                new_variables[key] == ""
            ):
                new_variables.pop(key)
        # Update the global variables to have access to the variables returned by the python file
        if "Variable Name" in self.returned_values_variables:
            for returned_values_variables in self.returned_values_variables[
                "Variable Name"
            ]:
                if returned_values_variables:
                    if (
                        returned_values_variables
                        not in variables_handling.loop_step_variables
                    ) and (
                        returned_values_variables
                        not in variables_handling.channels.keys()
                    ):
                        variables_handling.loop_step_variables.update(
                            {returned_values_variables: 0}
                        )


class Execute_Python_File_Config(Loop_Step_Config):
    """The configuration settings for the Execute_Python_File loop step."""

    def __init__(
        self,
        loop_step: Execute_Python_File,
        parent=None,
    ):
        super().__init__(parent, loop_step)
        self.loop_step = loop_step

        # Labels setup
        self.label_python_file_path = QLabel("Python File Path")
        self.label_python_exe_path = QLabel("Python Executable Path")
        self.python_packages_versions_label = QLabel("Python Packages and Versions")
        self.passing_variables_label = QLabel("Variables to pass to the Python file")
        self.read_variables_label = QLabel("Values returned by the Python file.")

        # Path_Button_Edit widgets setup
        self.path_button_edit_python_file = Path_Button_Edit()
        self.path_button_edit_python_exe = Path_Button_Edit()

        # Radio buttons setup
        self.radio_existing_env = QRadioButton("Use existing python environment")
        self.radio_specific_packages = QRadioButton("Use specific packages")
        self.radio_camels_python = QRadioButton("Use CAMELS python")
        # Button group to manage exclusive selection
        self.radio_button_group = QButtonGroup()
        self.radio_button_group.addButton(self.radio_existing_env)
        self.radio_button_group.addButton(self.radio_specific_packages)
        self.radio_button_group.addButton(self.radio_camels_python)
        self.radio_button_group.buttonToggled.connect(self.handle_radio_button_clicked)
        # Layout for radio buttons
        radio_layout = QHBoxLayout()
        radio_layout.addWidget(self.radio_camels_python)
        radio_layout.addWidget(self.radio_existing_env)
        radio_layout.addWidget(self.radio_specific_packages)

        # Connect radio button signals
        self.radio_existing_env.toggled.connect(self.toggle_add_remove_table)
        self.radio_specific_packages.toggled.connect(self.toggle_python_exe_path)

        # Horizontal line setup
        h_line = QFrame()
        h_line.setFrameShape(QFrame.HLine)
        h_line.setFrameShadow(QFrame.Sunken)

        # AddRemoveTable widget setup for python packages and their version when selecting the python packages explicitly
        self.add_remove_table_packages = AddRemoveTable(
            headerLabels=["Python Package", "Version"]
        )

        # AddRemoveTable for passing variables to the Python file
        # If no variable is given in the table, it will take the variable values from the running script
        self.add_remove_table_variables_passing = AddRemoveTable(
            headerLabels=["Variable Name", "Value"]
        )

        # AddRemoveTable for reading from the Python file
        # Value is either the index for the returned value if the file returns multiple things or the dictionary key if the file returns a dictionary
        self.add_remove_table_returned_values_variables = (
            AddRemoveTable_Returned_Variables(headerLabels=["Variable Name"])
        )

        # Tooltips
        self.label_python_file_path.setToolTip(
            "Specify the path to the Python file to be executed."
        )
        self.label_python_exe_path.setToolTip(
            "Specify the path to the Python executable (python.exe)."
        )
        self.path_button_edit_python_file.setToolTip(
            "Specify the path to the Python file to be executed."
        )
        self.path_button_edit_python_exe.setToolTip(
            "Specify the path to the Python executable (python.exe)."
        )
        self.radio_existing_env.setToolTip(
            "Run the Python file using the existing Python environment."
        )
        self.radio_specific_packages.setToolTip(
            "Create a virtual environment with specified packages to run the Python file."
        )
        self.radio_camels_python.setToolTip(
            "Run the Python file using the Python environment CAMELS is using."
        )
        self.add_remove_table_packages.setToolTip(
            "Add or remove Python packages and their versions for the virtual environment."
        )
        self.add_remove_table_variables_passing.setToolTip(
            "Specify variables and their values to be passed to the Python file during execution."
        )
        self.add_remove_table_returned_values_variables.setToolTip(
            "The Python file must return a dictionary.\nSpecify the keys of the dictionary that are returned from the Python file after execution."
        )

        # Initial paths setup
        self.path_button_edit_python_file.set_path(self.loop_step.file_path)
        self.path_button_edit_python_exe.set_path(self.loop_step.python_exe_path)
        self.path_button_edit_python_file.path_changed.connect(self.update_file_path)
        self.path_button_edit_python_exe.path_changed.connect(self.update_exe_path)

        # Save changes of the python packages table to the loop step
        self.add_remove_table_packages.table_model.itemChanged.connect(
            self.update_python_packages
        )
        self.add_remove_table_packages.change_table_data(
            self.loop_step.python_packages_versions
        )
        self.add_remove_table_packages.table_model.rowsRemoved.connect(
            self.on_rows_removed_packages
        )
        # Save changes of the variables passing table to the loop step
        self.add_remove_table_variables_passing.table_model.itemChanged.connect(
            self.update_variables_passing
        )
        self.add_remove_table_variables_passing.change_table_data(
            self.loop_step.variables_passing
        )
        self.add_remove_table_variables_passing.table_model.rowsRemoved.connect(
            self.on_rows_removed_passing
        )
        # Save changes of the returned variables table to the loop step
        self.add_remove_table_returned_values_variables.table_model.itemChanged.connect(
            self.update_returned_value_variables
        )
        self.add_remove_table_returned_values_variables.change_table_data(
            self.loop_step.returned_values_variables
        )
        self.add_remove_table_returned_values_variables.table_model.rowsRemoved.connect(
            self.on_rows_removed_returned
        )

        # Layout setup
        layout = self.layout()

        layout.addWidget(self.label_python_file_path, 1, 0)
        layout.addWidget(self.path_button_edit_python_file, 1, 1)

        # Create a widget to hold radio buttons and add it to the main layout
        radio_widget = QWidget()
        radio_widget.setLayout(radio_layout)
        layout.addWidget(radio_widget, 2, 0, 1, 2)

        layout.addWidget(self.label_python_exe_path, 3, 0)
        layout.addWidget(self.path_button_edit_python_exe, 3, 1)
        layout.addWidget(self.python_packages_versions_label, 4, 0, 1, 2)
        layout.addWidget(self.add_remove_table_packages, 5, 0, 1, 2)
        layout.addWidget(h_line, 6, 0, 1, 2)
        layout.addWidget(self.passing_variables_label, 7, 0, 1, 2)
        layout.addWidget(self.add_remove_table_variables_passing, 8, 0, 1, 2)
        layout.addWidget(self.read_variables_label, 9, 0, 1, 2)
        layout.addWidget(self.add_remove_table_returned_values_variables, 10, 0, 1, 2)

        # Set the selected radio button from the loop step
        if self.loop_step.radio_button_selected is not None:
            if (
                self.loop_step.radio_button_selected
                == "Use existing python environment"
            ):
                self.radio_existing_env.setChecked(True)
            elif self.loop_step.radio_button_selected == "Use specific packages":
                self.radio_specific_packages.setChecked(True)
            elif self.loop_step.radio_button_selected == "Use CAMELS python":
                self.radio_camels_python.setChecked(True)
        else:
            self.radio_camels_python.setChecked(True)

    def update_exe_path(self, path):
        self.loop_step.python_exe_path = path

    def update_file_path(self, path):
        self.loop_step.file_path = path

    def update_python_packages(self):
        self.loop_step.python_packages_versions = (
            self.add_remove_table_packages.update_table_data()
        )

    def update_variables_passing(self):
        self.loop_step.variables_passing = (
            self.add_remove_table_variables_passing.update_table_data()
        )

    def update_returned_value_variables(self):

        self.loop_step.returned_values_variables = (
            self.add_remove_table_returned_values_variables.update_table_data()
        )

    def handle_radio_button_clicked(self, button):
        self.loop_step.radio_button_selected = button.text()
        if button == self.radio_existing_env:
            # Handle selection for existing environment
            self.add_remove_table_packages.setVisible(False)
            self.path_button_edit_python_exe.setVisible(True)
            self.label_python_exe_path.setVisible(True)
            self.python_packages_versions_label.setVisible(False)
            self.loop_step.use_existing_env = True
            self.loop_step.use_specific_packages = False
            self.loop_step.use_camels_python = False
        elif button == self.radio_specific_packages:
            # Handle selection for specific packages
            self.add_remove_table_packages.setVisible(True)
            self.path_button_edit_python_exe.setVisible(False)
            self.label_python_exe_path.setVisible(False)
            self.python_packages_versions_label.setVisible(True)
            self.loop_step.use_specific_packages = True
            self.loop_step.use_existing_env = False
            self.loop_step.use_camels_python = False
        elif button == self.radio_camels_python:
            # Handle selection for CAMELS python
            # Hide AddRemoveTable and Python exe path
            self.add_remove_table_packages.setVisible(False)
            self.path_button_edit_python_exe.setVisible(False)
            self.label_python_exe_path.setVisible(False)
            self.python_packages_versions_label.setVisible(False)
            self.loop_step.use_camels_python = True
            self.loop_step.use_specific_packages = False
            self.loop_step.use_existing_env = False

    def toggle_add_remove_table(self, checked):
        if checked:
            self.add_remove_table_packages.setVisible(False)

    def toggle_python_exe_path(self, checked):
        if checked:
            self.path_button_edit_python_exe.setVisible(not checked)
            self.label_python_exe_path.setVisible(not checked)

    def on_rows_removed_returned(self, parent, first, last):
        self.loop_step.returned_values_variables = (
            self.add_remove_table_returned_values_variables.update_table_data()
        )

    def on_rows_removed_passing(self, parent, first, last):
        self.loop_step.variables_passing = (
            self.add_remove_table_variables_passing.update_table_data()
        )

    def on_rows_removed_packages(self, parent, first, last):
        self.loop_step.python_packages_versions = (
            self.add_remove_table_packages.update_table_data()
        )


class AddRemoveTable_Returned_Variables(AddRemoveTable):
    """Subclasses AddRemoveTable and changes the remove method to also remove the variable from the loop step variable dict."""

    def remove(self):
        try:
            index = self.table.selectedIndexes()[0]
        except IndexError:
            raise Exception("You need to select a row first!")
        row = index.row()
        variable_name_to_be_removed = self.tableData["Variable Name"][row]
        variables_handling.loop_step_variables.pop(variable_name_to_be_removed)
        super().remove()



from PySide6.QtWidgets import (
    QWidget,
    QTableWidgetItem,
    QLabel,
    QMessageBox,
    QLineEdit,
    QCheckBox,
    QGridLayout,
    QComboBox,
    QTextEdit,
    QSpacerItem,
    QSizePolicy,
    QPushButton,
    QDialog,
)
from PySide6.QtCore import Qt

import databroker

from nomad_camels.utility import databroker_export, variables_handling
from nomad_camels.ui_widgets import path_button_edit, warn_popup


class Datbroker_Exporter(QDialog):
    """ """

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowTitle("Databroker Export - NOMAD CAMELS")

        file_label = QLabel("Filename:")
        data_path = variables_handling.meas_files_path
        self.file_box = path_button_edit.Path_Button_Edit(
            self, default_dir=data_path, save_file=True, file_extension="*.nxs"
        )
        session_label = QLabel("Session name:")
        self.lineEdit_session = QLineEdit()

        self.checkBox_export_csv = QCheckBox("export data to .csv")
        self.checkBox_export_json = QCheckBox("export metadata to .json")

        catalog_label = QLabel("Databroker catalog name:")
        catalog_name = "CAMELS_CATALOG"
        if "databroker_catalog_name" in variables_handling.preferences:
            catalog_name = variables_handling.preferences["databroker_catalog_name"]
        catalogs = list(databroker.catalog)
        self.catalog_box = QComboBox()
        self.catalog_box.addItems(catalogs)
        if catalog_name in catalogs:
            self.catalog_box.setCurrentText(catalog_name)

        run_label = QLabel("Run:")
        self.run_box = QComboBox()

        label_name = QLabel("Plan name:")
        label_time = QLabel("Time of run:")
        label_exit = QLabel("Exit status:")
        label_user = QLabel("User name:")
        label_overview = QLabel("Protocol overview")
        label_description = QLabel("Description")
        self.label_name = QLineEdit()
        self.label_time = QLineEdit()
        self.label_exit = QLineEdit()
        self.label_user = QLineEdit()
        self.label_name.setEnabled(False)
        self.label_time.setEnabled(False)
        self.label_exit.setEnabled(False)
        self.label_user.setEnabled(False)
        self.textedit_overview = QTextEdit()
        self.textedit_description = QTextEdit()
        self.textedit_description.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard | Qt.TextSelectableByMouse
        )
        self.textedit_overview.setTextInteractionFlags(
            Qt.TextSelectableByKeyboard | Qt.TextSelectableByMouse
        )

        self.button_export = QPushButton("Export")
        self.button_cancel = QPushButton("Cancel")
        self.button_cancel.clicked.connect(self.close)
        self.button_export.clicked.connect(self.export_data)

        layout = QGridLayout()
        layout.addWidget(file_label, 0, 0)
        layout.addWidget(self.file_box, 0, 1)
        layout.addWidget(session_label, 1, 0)
        layout.addWidget(self.lineEdit_session, 1, 1)
        layout.addWidget(self.checkBox_export_csv, 2, 0)
        layout.addWidget(self.checkBox_export_json, 2, 1)
        left = QWidget()
        left.setLayout(layout)
        layout.addItem(
            QSpacerItem(
                0,
                0,
                hData=QSizePolicy.Policy.Minimum,
                vData=QSizePolicy.Policy.Expanding,
            ),
            6,
            3,
        )

        layout = QGridLayout()
        layout.addWidget(catalog_label, 0, 2)
        layout.addWidget(self.catalog_box, 0, 3)
        layout.addWidget(run_label, 1, 2)
        layout.addWidget(self.run_box, 1, 3)

        layout.addWidget(label_name, 2, 2)
        layout.addWidget(label_time, 3, 2)
        layout.addWidget(label_exit, 4, 2)
        layout.addWidget(label_user, 5, 2)
        layout.addWidget(self.label_name, 2, 3)
        layout.addWidget(self.label_time, 3, 3)
        layout.addWidget(self.label_exit, 4, 3)
        layout.addWidget(self.label_user, 5, 3)
        layout.addWidget(label_description, 6, 2, 1, 2)
        layout.addWidget(self.textedit_description, 7, 2, 1, 2)
        mid = QWidget()
        mid.setLayout(layout)

        layout = QGridLayout()
        # layout.addWidget(label_description, 0, 4)
        # layout.addWidget(self.textedit_description, 1, 4, 1, 2)
        layout.addWidget(label_overview, 3, 4)
        layout.addWidget(self.textedit_overview, 4, 4, 1, 2)
        right = QWidget()
        right.setLayout(layout)

        layout = QGridLayout()
        layout.addWidget(left, 1, 0, 1, 3)
        layout.addWidget(mid, 0, 0)
        layout.addWidget(right, 0, 1, 1, 2)
        layout.addWidget(self.button_export, 5, 1)
        layout.addWidget(self.button_cancel, 5, 2)
        self.setLayout(layout)
        self.show()

        self.change_catalog()
        self.catalog_box.currentTextChanged.connect(self.change_catalog)
        self.change_run()
        self.run_box.currentTextChanged.connect(self.change_run)
        self.adjustSize()

    def export_data(self):
        self.setCursor(Qt.CursorShape.WaitCursor)
        filename = self.file_box.get_path()
        run = self.run_box.currentData()
        session = self.lineEdit_session.text()
        exp_csv = self.checkBox_export_csv.isChecked()
        exp_json = self.checkBox_export_json.isChecked()
        catalog = self.catalog_box.currentText()
        databroker_export.export_run(
            filename=filename,
            run_number=run,
            session_name=session,
            export_to_csv=exp_csv,
            export_to_json=exp_json,
            catalog_name=catalog,
        )
        self.setCursor(Qt.CursorShape.ArrowCursor)
        warn_popup.WarnPopup(self, "Successfully exported data!", "Data exported", True)
        self.accept()

    def change_catalog(self):
        self.setCursor(Qt.WaitCursor)
        catalog_name = self.catalog_box.currentText()
        if not catalog_name:
            return
        catalog = databroker.catalog[catalog_name]
        self.run_box.clear()
        run_time_names = []
        for run_name in list(catalog):
            run = catalog[run_name]
            run_time = run.metadata["start"]["time"]
            run_time = databroker_export.timestamp_to_ISO8601(run_time)
            run_time_names.append((run_time, run_name))
        for run in sorted(run_time_names, reverse=True):
            run_time, run_name = run
            self.run_box.addItem(f"{run_time} - {run_name}", userData=run_name)
        self.setCursor(Qt.ArrowCursor)

    def change_run(self):
        run_name = self.run_box.currentData()
        catalog_name = self.catalog_box.currentText()
        if run_name:
            catalog = databroker.catalog[catalog_name]
            run = catalog[run_name]
            metadata = run.metadata["start"]
            stopdata = run.metadata["stop"]
        else:
            metadata = {}
            stopdata = {}
        time = ""
        if "time" in metadata:
            time = databroker_export.timestamp_to_ISO8601(metadata["time"])
        self.label_time.setText(time)
        plan_name = ""
        if "plan_name" in metadata:
            plan_name = metadata["plan_name"]
        self.label_name.setText(plan_name)
        description = ""
        if "description" in metadata:
            description = metadata["description"]
        self.textedit_description.setText(description)
        user = ""
        if "user" in metadata:
            user = metadata["user"]["name"]
        self.label_user.setText(user)
        protocol_overview = ""
        if "protocol_overview" in metadata:
            protocol_overview = metadata["protocol_overview"]
        self.textedit_overview.setText(protocol_overview)
        exit_status = ""
        if stopdata and "exit_status" in stopdata:
            exit_status = stopdata["exit_status"]
        self.label_exit.setText(exit_status)

    def closeEvent(self, a0) -> None:
        discard_dialog = QMessageBox.question(
            self,
            "Discard Changes?",
            f"All changes will be lost!",
            QMessageBox.Yes | QMessageBox.No,
        )
        if discard_dialog != QMessageBox.Yes:
            a0.ignore()
            return
        super().closeEvent(a0)


if __name__ == "__main__":
    from PySide6.QtWidgets import QApplication

    app = QApplication([])

    widge = Datbroker_Exporter()
    widge.show()
    app.exec()



from PySide6.QtWidgets import QWidget, QLabel, QGridLayout

from nomad_camels.main_classes.loop_step import Loop_Step_Config, Loop_Step
from nomad_camels.ui_widgets.variable_tool_tip_box import Variable_Box


class Export_Data_Step(Loop_Step):
    """
    A loopstep to export the measurement data at fixed points. Convenient to
    already start evaluation during long measurements.
    """

    def __init__(self, name="", parent_step=None, step_info=None, **kwargs):
        super().__init__(name, parent_step, step_info, **kwargs)
        self.step_type = "Export Data"

    def get_protocol_string(self, n_tabs=1):
        """The protocol just calls `bps.wait(`wait_time`)`, where `wait_time` is
        evaluated by the protocol's evaluator."""
        tabs = "\t" * n_tabs
        protocol_string = super().get_protocol_string(n_tabs)
        protocol_string += f"{tabs}if uids:\n"
        protocol_string += f"{tabs}\truns uids:\n"
        protocol_string += (
            f"{tabs}broker_to_NX(runs, save_path, plots,"
            "session_name=session_name,"
            "export_to_csv=export_to_csv,"
            "export_to_json=export_to_json)\n"
        )
        protocol_string += f'{tabs}yield from bps.sleep(eva.eval("{self.wait_time}"))\n'
        return protocol_string


class Export_Data_Step_Config(Loop_Step_Config):
    """Displaying only what the step does."""

    def __init__(self, loop_step: Export_Data_Step, parent=None):
        super().__init__(parent, loop_step)
        self.sub_widget = QLabel(
            "This step exports the current measured data to"
            "the same directory where the final data will be"
            'saved. The filename will include "_intermediate"'
            "at the end."
        )
        self.layout().addWidget(self.sub_widget, 1, 0)



from ophyd import SignalRO
import numpy as np
from collections import namedtuple


class Variable_Signal(SignalRO):
    def __init__(
        self,
        name,
        value=None,
        timestamp=None,
        parent=None,
        labels=None,
        kind="normal",
        tolerance=None,
        rtolerance=None,
        metadata=None,
        cl=None,
        attr_name="",
        variables_dict=None,
    ):
        super().__init__(
            name=name,
            value=value,
            timestamp=timestamp,
            parent=parent,
            labels=labels,
            kind=kind,
            tolerance=tolerance,
            rtolerance=rtolerance,
            metadata=metadata,
            cl=cl,
            attr_name=attr_name,
        )
        self.variables_dict = variables_dict or {}
        self.vars = list(sorted(self.variables_dict.keys()))

    def describe(self):
        info = super().describe()
        info[self.name]["source"] = "local_NOMAD_CAMELS_variables"
        # self.vars = list(sorted(self.variables_dict.keys()))
        info[self.name]["variables"] = self.vars
        return info

    def get(self):
        # data = []
        # for key in sorted(self.variables_dict.keys()):
        #     if key not in self.vars:
        #         continue
        #     data.append(self.variables_dict[key])
        # self._readback = np.asarray(data)
        data = {}
        for key in sorted(self.variables_dict.keys()):
            if key not in self.vars:
                continue
            data[key] = self.variables_dict[key]
        dev_tuple = namedtuple(self.name, data.keys())
        self._readback = dev_tuple(**data)
        return super().get()
